/*****************************************************************************\
 *                                                                           *
 * TestMethods.cpp                                                           *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
 * This test just makes sure that the constructor, destructor, method        *
 * invoker, getter, and setter methods generated by xrttigen work properly   *
 *                                                                           *
 \****************************************************************************/

#include <Essential/Util.h>
#include <Xrtti/Xrtti.h>
#include "test/TestMethods.h"

using namespace Xrtti;

static const Method *LookupMethod(const Class &classRef, const char *name)
{
    Xrtti::U32 count = classRef.GetMethodCount();

    for (Xrtti::U32 i = 0; i < count; i++) {
        const Method &method = classRef.GetMethod(i);
        if (String::Equal(method.GetName(), name)) {
            return &method;
        }
    }

    Log::ErrorF("TestMethods has no %s method\n", name);
    Util::Terminate();

    return 0;
}


static const Field *LookupField(const Class &classRef, const char *name)
{
    Xrtti::U32 count = classRef.GetFieldCount();

    for (Xrtti::U32 i = 0; i < count; i++) {
        const Field &field = classRef.GetField(i);
        if (String::Equal(field.GetName(), name)) {
            return &field;
        }
    }

    Log::ErrorF("TestMethods has no %s field\n", name);
    Util::Terminate();

    return 0;
}


int main(int argc, char **argv)
{
    const Context *pContext = LookupContext("TestMethods");

    if (!pContext) {
        Log::ErrorF("Failed to lookup TestMethods Context\n");
        Util::Terminate();
    }

    if (pContext->GetType() != Context::Type_Class) {
        Log::ErrorF("TestMethods Context is not a class\n");
        Util::Terminate();
    }

    const Class &classRef = * (const Class *) pContext;

    // Create one
    TestMethods *pCreated = (TestMethods *) classRef.Create();
    if (!pCreated) {
        Log::ErrorF("TestMethods default constructor returned NULL\n");
        Util::Terminate();
    }

    // Invoke "Identify" method
    Value returnValue;
    LookupMethod(classRef, "Identify")->Invoke(pCreated, returnValue, NULL);

    // Invoke "Count" method repeatedly
    for (Xrtti::U32 i = 0; i < 10; i++) {
        LookupMethod(classRef, "Count")->Invoke(pCreated, returnValue, NULL);
        Log::MessageF("TestMethods Count: %d\n", returnValue.short_v);
    }
    
    // Invoke "Sum" method
    Value args[2];
    args[0].int_v = 1;
    args[1].int_v = 2;
    LookupMethod(classRef, "Sum")->Invoke(pCreated, returnValue, args);
    Log::MessageF("TestMethods Sum: %d\n", returnValue.int_v);

    // Invoke "Allocate" method
    Value arg;
    arg.unsigned_long_v = 5;
    LookupMethod(classRef, "Allocate")->Invoke(pCreated, returnValue, &arg);
    Xrtti::S32 *pAllocated = (Xrtti::S32 *) returnValue.pointer_v;
    for (Xrtti::U32 i = 0; i < arg.unsigned_long_v; i++) {
        Log::MessageF("TestMethods Allocate: %d\n", pAllocated[i]);
    }
    delete [] pAllocated;

    // Invoke "GetInnerRef" method
    LookupMethod(classRef, "GetInnerRef")->Invoke(pCreated, returnValue, NULL);
    const TestMethods::Inner &inner = 
        * (TestMethods::Inner *) returnValue.pointer_v;
    Log::MessageF("TestMethods::GetInnerRef a: %u, b: %f\n", inner.a, inner.b);

    // Invoke "GetInnerCopy" method
    TestMethods::Inner innerCopy;
    returnValue.pointer_v = &innerCopy;
    LookupMethod(classRef, "GetInnerCopy")->Invoke(pCreated, returnValue, NULL);
    Log::MessageF("TestMethods::GetInnerCopy a: %u, b: %f\n", innerCopy.a, 
                  innerCopy.b);

    // Invoke "TestParams" method
    Value args2[6];
    args2[0].int_v = 1;
    args2[1].float_v = 2.0;
    args2[2].pointer_v = (void *) "\"3\"";
    TestMethods::Inner innerD;
    innerD.a = 4, innerD.b = 5.0;
    args2[3].pointer_v = &innerD;
    TestMethods::Inner innerE;
    innerE.a = 6, innerE.b = 7.0;
    args2[4].pointer_v = &innerE;
    TestMethods::Inner innerF;
    innerF.a = 8, innerF.b = 9.0;
    args2[5].pointer_v = &innerF;
    LookupMethod(classRef, "TestParams")->Invoke(pCreated, returnValue, args2);

    // Set "counterM" parameter
    Value counter;
    counter.short_v = 123;
    LookupField(classRef, "counterM")->Set(pCreated, counter);
    counter.short_v = 0;

    // Get "counterM" parameter
    LookupField(classRef, "counterM")->Get(pCreated, counter);
    Log::MessageF("TestMethods::Get counterM: %d\n", counter.short_v);

    // Set "innerM" parameter
    TestMethods::Inner innerX;
    innerX.a = 901, innerX.b = 902.0;
    Value innerArg;
    innerArg.pointer_v = &innerX;
    LookupField(classRef, "innerM")->Set(pCreated, innerArg);
    innerX.a = 0, innerX.b = 0.0;

    // Get "innerM" parameter
    LookupField(classRef, "innerM")->Get(pCreated, innerArg);
    Log::MessageF("TestMethods::Get innerM a: %d, b: %f\n", 
                  innerX.a, innerX.b);

    // Set "pInnerM" parameter
    LookupField(classRef, "pInnerM")->Set(pCreated, innerArg);
    innerArg.pointer_v = 0;

    // Get "pInnerM" parameter
    LookupField(classRef, "pInnerM")->Get(pCreated, innerArg);
    Log::MessageF("TestMethods::Get pInnerM a: %d, b: %f\n", 
                  ((TestMethods::Inner *) innerArg.pointer_v)->a,
                  ((TestMethods::Inner *) innerArg.pointer_v)->b);

    // Delete it
    classRef.Delete(pCreated);

    // Create array
    TestMethods *pResult = (TestMethods *) classRef.CreateArray(10);

    for (Xrtti::U32 i = 0; i < 10; i++) {
        // Invoke "indentifer" method
        LookupMethod(classRef, "Identify")->Invoke
            (&(pResult[i]), returnValue, NULL);
    }

    // Delete them
    classRef.DeleteArray((void **) pResult);

    ShutDown();

    return 0;
}
