/*****************************************************************************\
 *                                                                           *
 * XrttiToCpp.cpp                                                            *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include <Essential/Hashtable.h>
#include <Essential/Vector.h>
#include <Xrtti/Xrtti.h>
#include "test/XrttiToCpp.h"

using namespace Essential;
using namespace Xrtti;

typedef Xrtti::U32 X_U32;
typedef Xrtti::Enumeration X_Enumeration;

enum Access
{
    Unknown,
    Public,
    Protected,
    Private
};

class OutputIndentStream
{
public:

    OutputIndentStream(OutputStream &os)
        : osM(os), indentM(0), newlineM(true)
    {
    }

    void indent()
    {
        indentM++;
    }

    void deindent()
    {
        indentM--;
    }

    inline bool Write(const String &string)
    {
        if (!this->WriteIndentIfNecessary()) {
            return false;
        }

        return osM.WriteA(string);
    }

    inline bool WriteF(const char *pFormat, ...)
    {
        va_list valist;
        
        va_start(valist, pFormat);

        bool ret = this->Write(String::FormatV(pFormat, valist));
        
        va_end(valist);

        return ret;
    }

    inline bool WriteLine()
    {
        if (!this->Write("\n")) {
            return false;
        }

        newlineM = true;

        return true;
    }

    inline bool WriteLine(const String &string)
    {
        if (!this->Write(string)) {
            return false;
        }

        return this->WriteLine();
    }

    inline bool WriteLineF(const char *pFormat, ...)
    {
        va_list valist;
        
        va_start(valist, pFormat);

        bool ret = this->WriteLine(String::FormatV(pFormat, valist));
        
        va_end(valist);

        return ret;
    }
    
private:

    inline bool WriteIndentIfNecessary()
    {
        if (!newlineM) {
            return true;
        }

        newlineM = false;
        
        for (X_U32 i = 0; i < indentM; i++) {
            if (!osM.WriteA("    ")) {
                return false;
            }
        }

        return true;
    }

    OutputStream &osM;
    X_U32 indentM;
    bool newlineM;
};


static void SortIntoByName(const Context &context,
                            Vector<const Context *> &vContexts);
static void AddToHierarchy(const Context &context);
static void CollectEnumerations(const Context &context);
static bool IsInAnonymousChain(const Structure &structure);
static void PrintIncomplete(OutputIndentStream &out, 
                            const Incomplete &incomplete);
static Access PrintAccess(OutputIndentStream &out, AccessType accessType,
                          Access access);
static void PrintFunctionTypedefs(OutputIndentStream &out, 
                                  const Context &context);
static void PrintChildrenContexts(OutputIndentStream &out, 
                                  const Context &context);
static void PrintNamespace(OutputIndentStream &out,
                           const Namespace &namespaceRef);
static void PrintStructure(OutputIndentStream &out, const Structure &structRef);
static void PrintContext(OutputIndentStream &out, const Context &context);
static void PrintField(OutputIndentStream &out, const Field &field);
static void PrintConstructorSuffix(OutputIndentStream &out, 
                                   const Constructor &c);
static void PrintMethod(OutputIndentStream &out, const Method &method);
static void PrintEnumeration(OutputIndentStream &out,
                             const X_Enumeration &enumeration);
static void PrintValue(OutputIndentStream &out, const Type &type,
                       const Value &value);
static void PrintTypePrefix(OutputIndentStream &out, const Type &type);
static void PrintTypeSuffix(OutputIndentStream &out, const Type &type);
static void PrintConstructorSignatureSuffix(OutputIndentStream &out, 
                                            const ConstructorSignature &sig,
                                            Vector<const char *> *pArgNames);

class ContextExtra
{
public:

    static ContextExtra &Get(const Context &context)
    {
        ContextExtra *pContextExtra;
        
        if (htContextExtrasG.Get(String(context.GetFullName()),
                                 /* returns */ pContextExtra)) {
            return *pContextExtra;
        }

        pContextExtra = new ContextExtra();

        ContextExtra *pJunk;

        htContextExtrasG.Put(context.GetFullName(), pContextExtra, pJunk);

        return *pContextExtra;
    }

    bool HasBeenPrinted()
    {
        return hasBeenPrintedM;
    }

    void SetHasBeenPrinted()
    {
        hasBeenPrintedM = true;
    }

    void AddChild(const Context &context)
    {
        const Context *pContext = &context;

        vChildrenM.Append(pContext);
    }

    X_U32 GetChildCount()
    {
        return vChildrenM.Count();
    }

    const Context &GetChild(X_U32 index)
    {
        return *(vChildrenM[index]);
    }

    void AddEnumeration(const X_Enumeration &enumeration)
    {
        const X_Enumeration *pEnumeration = &enumeration;

        // Make sure it's not there already
        if (vEnumerationsM.Contains(pEnumeration)) {
            return;
        }

        vEnumerationsM.Append(pEnumeration);
    }

    X_U32 GetEnumerationCount()
    {
        return vEnumerationsM.Count();
    }

    const X_Enumeration &GetEnumeration(X_U32 index)
    {
        return *(vEnumerationsM[index]);
    }

    void SortChildren()
    {
        X_U32 count = vChildrenM.Count();

        // This will hold the sorted children
        Vector<const Context *> vSortedChildren;

        // For each child
        for (X_U32 i = 0; i < count; i++) {
            const Context *pChild = vChildrenM[i];

            // For each child already in the sorted list, if this one should
            // go before it, insert it there
            X_U32 where;
            for (where = 0; where < i; where++) {
                if (ShouldComeBefore(pChild, vSortedChildren[where])) {
                    break;
                }
            }

            vSortedChildren.Insert(where, pChild);
        }

        // Replace the children with the sorted children
        vChildrenM.Clear();
        vChildrenM.Append(vSortedChildren);

        // Now sort the children of this
        for (X_U32 i = 0; i < count; i++) {
            Get(*vChildrenM[i]).SortChildren();
        }
    }

private:

    class AutoDeleter
    {
    public:

        ~AutoDeleter()
        {
            Hashtable__ValueEnumeration<String, ContextExtra *> 
                values(htContextExtrasG);
            
            while (values.HasElements()) {
                ContextExtra *pCe = values.GetNextElement();

                delete pCe;
            }

            htContextExtrasG.Clear();
        }

        static AutoDeleter autoDeleterG;
    };

    ContextExtra()
        : hasBeenPrintedM(false)
    {
    }

    // c1 should come before c2 if:
    // a. Any of c2's bases or parent is c1
    // b. Recursively, any of c2's children's bases or parent is c1
    // c. Recursively, any of (a) or (b) is true for c2 and any of c1's
    //    children
    bool ShouldComeBefore(const Context *pC1, const Context *pC2)
    {
        // If pC2 has as a parent pC1, then return true
        const Context *pParent = pC2->GetParent();
        if (pC1 == pParent) {
            return true;
        }
        
        // If pC2 has as a base pC1, then return true
        if ((pC2->GetType() == Context::Type_Class) ||
            (pC2->GetType() == Context::Type_Class) ||
            (pC2->GetType() == Context::Type_Union)) {
            const Structure *pS2 = (const Structure *) pC2;
            X_U32 count = pS2->GetBaseCount();

            for (X_U32 i = 0; i < count; i++) {
                const Structure &baseStruct = pS2->GetBase(i).GetStructure();
                if (pC1 == ((const Context *) &baseStruct)) {
                    return true;
                }
            }
        }

        // Recursively for pC1 and all children of pC2
        {
            ContextExtra &ce = ContextExtra::Get(*pC2);

            X_U32 count = ce.GetChildCount();
        
            for (X_U32 i = 0; i < count; i++) {
                const Context &child = ce.GetChild(i);
                if (ShouldComeBefore(pC1, &child)) {
                    return true;
                }
            }
        }

        // Recursively for pC2 and all children of pC1
        {
            ContextExtra &ce = ContextExtra::Get(*pC1);
            
            X_U32 count = ce.GetChildCount();
            
            for (X_U32 i = 0; i < count; i++) {
                const Context &child = ce.GetChild(i);
                if (ShouldComeBefore(&child, pC2)) {
                    return true;
                }
            }
        }

        return false;
    }

    Vector<const Context *> vChildrenM;
    Vector<const X_Enumeration *> vEnumerationsM;
    bool hasBeenPrintedM;

    static Hashtable<String, ContextExtra *> htContextExtrasG;
};

Hashtable<String, ContextExtra *> ContextExtra::htContextExtrasG;
ContextExtra::AutoDeleter ContextExtra::AutoDeleter::autoDeleterG;

class FunctionExtra
{
public:
    
    static FunctionExtra &Get(const MethodSignature &sig)
    {
        FunctionExtra *pFunctionExtra;
        
        if (htFunctionExtrasG.Get(&sig, /* returns */ pFunctionExtra)) {
            return *pFunctionExtra;
        }

        pFunctionExtra = new FunctionExtra();

        FunctionExtra *pJunk;

        htFunctionExtrasG.Put(&sig, pFunctionExtra, pJunk);

        return *pFunctionExtra;
    }

    X_U32 GetNumber()
    {
        return numberM;
    }

    bool HasBeenPrinted()
    {
        return hasBeenPrintedM;
    }

    void SetHasBeenPrinted()
    {
        hasBeenPrintedM = true;
    }

private:

    class AutoDeleter
    {
    public:

        ~AutoDeleter()
        {
            Hashtable__ValueEnumeration<const MethodSignature *,
                                        FunctionExtra *> 
                values(htFunctionExtrasG);
            
            while (values.HasElements()) {
                FunctionExtra *pFe = values.GetNextElement();

                delete pFe;
            }

            htFunctionExtrasG.Clear();
        }

        static AutoDeleter autoDeleterG;
    };

    FunctionExtra()
        : numberM(nextNumberG++), hasBeenPrintedM(false)
    {
    }

    X_U32 numberM;
    bool hasBeenPrintedM;

    static X_U32 nextNumberG;
    static Hashtable<const MethodSignature *, 
                     FunctionExtra *> htFunctionExtrasG;

};

X_U32 FunctionExtra::nextNumberG = 1;
Hashtable<const MethodSignature *, 
          FunctionExtra *> FunctionExtra::htFunctionExtrasG;
FunctionExtra::AutoDeleter FunctionExtra::AutoDeleter::autoDeleterG;

static void SortIntoByName(const Context &context, 
                           Vector<const Context *> &vContexts)
{
    const Context *pContext = &context;

    // Find where the context goes
    X_U32 count = vContexts.Count();

    for (X_U32 i = 0; i < count; i++) {
        const Context *pExisting = vContexts[i];

        // If pExisting comes after pContext, then insert pContext here
        if (String(pExisting->GetName()) > String(pContext->GetName())) {
            vContexts.Insert(i, pContext);
            return;
        }
    }

    // Must go at the end
    vContexts.Append(pContext);
}

static void AddToHierarchy(const Context &context)
{
    // Add this to the list of children of the parent context
    const Context *pParent = context.GetParent();

    // Ignore the root context
    if (!pParent) {
        return;
    }

    ContextExtra &ceParent = ContextExtra::Get(*pParent);

    ceParent.AddChild(context);
}

static void CollectEnumerations(const Context &context)
{
    ContextExtra &ce = ContextExtra::Get(context);

    // Do the sub-contexts
    X_U32 count = ce.GetChildCount();

    for (X_U32 i = 0; i < count; i++) {
        CollectEnumerations(ce.GetChild(i));
    }

    // Now look at each field of this context, if it has fields, and
    // see if there is an enumeration type to collect
    if ((context.GetType() != Context::Type_Class) &&
        (context.GetType() != Context::Type_Struct) &&
        (context.GetType() != Context::Type_Union)) {
        return;
    }

    const Structure &structure = (const Structure &) context;

    count = structure.GetFieldCount();

    for (X_U32 i = 0; i < count; i++) {
        const Type &type = structure.GetField(i).GetType();

        if (type.GetBaseType() != Type::BaseType_Enumeration) {
            continue;
        }

        const X_Enumeration &enumeration = ((const TypeEnumeration &) type).
            GetEnumeration();

        ContextExtra &ceEnum = ContextExtra::Get(enumeration.GetContext());

        ceEnum.AddEnumeration(enumeration);
    }
}

static bool IsInAnonymousChain(const Structure &structure)
{
    if (structure.IsAnonymous()) {
        return true;
    }

    const Context *pParent = structure.GetParent();

    if (pParent == NULL) {
        return false;
    }

    if ((pParent->GetType() != Context::Type_Class) &&
        (pParent->GetType() != Context::Type_Struct) &&
        (pParent->GetType() != Context::Type_Union)) {
        return false;
    }
    
    const Structure &parentStructure = (const Structure &) *pParent;

    return IsInAnonymousChain(parentStructure);
}

static void PrintIncomplete(OutputIndentStream &out,
                            const Incomplete &incomplete)
{
    out.WriteLineF("class %s;\n", incomplete.GetName());
}

static Access PrintAccess(OutputIndentStream &out, AccessType accessType,
                          Access access)
{
    switch (accessType) {
    case AccessType_Public:
        if (access != Public) {
            out.WriteLine("public:");
            return Public;
        }
        break;
    case AccessType_Protected:
        if (access != Protected) {
            out.WriteLine("protected:");
            return Protected;
        }
        break;
    default: // AccessType_Private
        if (access != Private) {
            out.WriteLine("private:");
            return Private;
        }
        break;
    }

    return access;
}

static void PrintFunctionTypedefs(OutputIndentStream &out,
                                  const Context &context)
{
    ContextExtra &ce = ContextExtra::Get(context);

    // Do the sub-contexts
    X_U32 count = ce.GetChildCount();

    for (X_U32 i = 0; i < count; i++) {
        PrintFunctionTypedefs(out, ce.GetChild(i));
    }

    // Non-Structure Contexts cannot have fields, so skip them
    if ((context.GetType() != Context::Type_Class) &&
        (context.GetType() != Context::Type_Struct) &&
        (context.GetType() != Context::Type_Union)) {
        return;
    }

    const Structure &structure = (const Structure &) context;

    count = structure.GetFieldCount();

    for (X_U32 i = 0; i < count; i++) {
        const Type &type = structure.GetField(i).GetType();

        if (type.GetBaseType() != Type::BaseType_Function) {
            continue;
        }

        const MethodSignature &sig = 
            ((const TypeFunction &) type).GetSignature();

        FunctionExtra &fe = FunctionExtra::Get(sig);

        if (fe.HasBeenPrinted()) {
            continue;
        }

        // Print the typedef of it
        out.Write("typedef ");
        PrintTypePrefix(out, sig.GetReturnType());
        out.WriteF(" (function_%lu) ", fe.GetNumber());
        PrintConstructorSignatureSuffix(out, sig, 0);
        out.WriteLine(";");
    }
}

static void PrintChildrenContexts(OutputIndentStream &out, 
                                  const Context &context)
{
    ContextExtra &ce = ContextExtra::Get(context);

    X_U32 count = ce.GetChildCount();

    for (X_U32 i = 0; i < count; i++) {
        PrintContext(out, ce.GetChild(i));
    }
}

static void PrintNamespace(OutputIndentStream &out,
                           const Namespace &namespaceRef)
{
    // If the parent is a namespace, print it first
    const Context *pParent = namespaceRef.GetParent();
    if (pParent) {
        PrintContext(out, *pParent);
    }

    ContextExtra &ce = ContextExtra::Get(namespaceRef);

    if (ce.HasBeenPrinted()) {
        return;
    }

    ce.SetHasBeenPrinted();

    // Skip it if it's the root namespace
    if (!namespaceRef.GetParent()) {
        return;
    }

    out.WriteLineF("namespace %s", namespaceRef.GetName());

    out.WriteLineF("{");

    out.indent();

    // Print the enumerations
    X_U32 count = ce.GetEnumerationCount();
    
    for (X_U32 i = 0; i < count; i++) {
        PrintEnumeration(out, ce.GetEnumeration(i));
    }

    count = ce.GetChildCount();

    for (X_U32 i = 0; i < count; i++) {
        PrintContext(out, ce.GetChild(i));
    }

    out.deindent();

    out.WriteLine("}");
    out.WriteLine();
}

// This is only called on anonymous structs to in-place print out types
static void PrintStructure(OutputIndentStream &out,
                           const Structure &structureRef)
{
    // If the parent is a namespace, print it first
    const Context *pParent = structureRef.GetParent();
    if (pParent) {
        PrintContext(out, *pParent);
    }

    ContextExtra &ce = ContextExtra::Get(structureRef);

    if (ce.HasBeenPrinted()) {
        return;
    }

    ce.SetHasBeenPrinted();

    Access access;

    if (structureRef.IsAnonymous()) {
        if (structureRef.GetType() == Context::Type_Class) {
            access = Private;
            out.Write("class");
        }
        else if (structureRef.GetType() == Context::Type_Struct) {
            access = Public;
            out.Write("struct");
        }
        else { // Context::Type_Union
            access = Public;
            out.Write("union");
        }
    }
    else {
        if (structureRef.GetType() == Context::Type_Class) {
            access = Private;
            out.WriteF("class %s", structureRef.GetName());
        }
        else if (structureRef.GetType() == Context::Type_Struct) {
            access = Public;
            out.WriteF("typedef struct %s", structureRef.GetName());
        }
        else { // Context::Type_Union
            access = Public;
            out.WriteF("typedef union %s", structureRef.GetName());
        }
    }

    // bases
    X_U32 count = structureRef.GetBaseCount();

    if (count) {
        out.Write(" : ");
        for (X_U32 i = 0; i < count; i++) {
            // There is no support for saying the access type of the
            // inheritence ...
            out.WriteF("public %s", 
                       structureRef.GetBase(i).GetStructure().GetName());
            if (i < (count - 1)) {
                out.Write(", ");
            }
        }
    }
    
    out.WriteLine();

    out.WriteLine("{");

    out.indent();

    // Print the enumerations
    count = ce.GetEnumerationCount();
    
    for (X_U32 i = 0; i < count; i++) {
        const X_Enumeration &enumeration = ce.GetEnumeration(i);
        access = PrintAccess(out, enumeration.GetAccessType(), access);
        PrintEnumeration(out, enumeration);
    }

    out.deindent();

    count = ce.GetChildCount();

    for (X_U32 i = 0; i < count; i++) {
        const Context &child = ce.GetChild(i);

        if ((child.GetType() == Context::Type_Class) ||
            (child.GetType() == Context::Type_Struct) ||
            (child.GetType() == Context::Type_Union)) {
            const Structure &structure = * (const Structure *) & child;

            // If it's anonymous, don't print it out; it will be printed
            // out with the field that it defines
            if (structure.IsAnonymous()) {
                continue;
            }
            
            access = PrintAccess(out, structure.GetAccessType(), access);

            out.WriteLine();
        } 
        
        out.indent();
    
        PrintContext(out, child);

        out.deindent();
    }

    out.indent();

    if (structureRef.GetType() != Context::Type_Union) {
        const Struct &structRef = (const Struct &) structureRef;

        // Now print the constructors and destructor of the struct - but only
        // if this is not anonymous
        if (!IsInAnonymousChain(structureRef)) {
            count = structRef.GetConstructorCount();

            for (X_U32 i = 0; i < count; i++) {
                const Constructor &constructor = structRef.GetConstructor(i);
                access = PrintAccess(out, constructor.GetAccessType(), access);
                out.WriteF("%s", structureRef.GetName());
                PrintConstructorSuffix(out, constructor);
                if (i < (count - 1)) {
                    out.WriteLine();
                }
            }

            if (structRef.HasDestructor()) {
                const Destructor &destructor = structRef.GetDestructor();
                access = PrintAccess(out, destructor.GetAccessType(), access);
                if (destructor.IsVirtual()) {
                    out.WriteF("virtual ");
                }
                out.WriteF("~%s()", structureRef.GetName());
                if (destructor.IsPureVirtual()) {
                    out.WriteF(" = 0");
                }
                out.WriteLineF(";");
            }
        }

        // Now print the methods of the struct
        count = structRef.GetMethodCount();

        for (X_U32 i = 0; i < count; i++) {
            const Method &method = structRef.GetMethod(i);
            access = PrintAccess(out, method.GetAccessType(), access);
            PrintMethod(out, method);
            if (i < (count - 1)) {
                out.WriteLine();
            }
        }
    }
    
    // Now print the fields of the structure
    count = structureRef.GetFieldCount();

    for (X_U32 i = 0; i < count; i++) {
        const Field &field = structureRef.GetField(i);
        access = PrintAccess(out, field.GetAccessType(), access);
        PrintField(out, field);
        if (i < (count - 1)) {
            out.WriteLine();
        }
    }

    out.deindent();

    if (structureRef.IsAnonymous()) {
        out.Write("}");
    }
    else {
        if (structureRef.GetType() == Context::Type_Class) {
            out.WriteLine("};");
        }
        else {
            out.WriteLineF("} %s;", structureRef.GetName());
        }
        out.WriteLine();
    }
}

static void PrintContext(OutputIndentStream &out, const Context &context)
{
    switch (context.GetType()) {
    case Context::Type_Class:
        // Don't print out anonymous structs - they will be printed with the
        // field itself
        if (!((const Class &) context).IsAnonymous()) {
            PrintStructure(out, (const Class &) context);
        }
        break;
    case Context::Type_Incomplete:
        PrintIncomplete(out, (const Incomplete &) context);
        break;
    case Context::Type_Namespace:
        PrintNamespace(out, (const Namespace &) context);
        break;
    case Context::Type_Struct:
        // Don't print out anonymous structs - they will be printed with the
        // field itself
        if (!((const Struct &) context).IsAnonymous()) {
            PrintStructure(out, (const Struct &) context);
        }
        break;
    case Context::Type_Union:
        // Don't print out anonymous unions - they will be printed with the
        // field itself
        if (!((const Union &) context).IsAnonymous()) {
            PrintStructure(out, (const Union &) context);
        }
        break;
    }
}

static void PrintField(OutputIndentStream &out, const Field &field)
{
    if (field.IsStatic()) {
        out.Write("static ");
    }

    // Print the type prefix
    PrintTypePrefix(out, field.GetType());

    if (*(field.GetName())) {
        out.WriteF(" %s", field.GetName());
    }
    
    PrintTypeSuffix(out, field.GetType());

    out.WriteLine(";");
}

static void PrintConstructorSuffix(OutputIndentStream &out, 
                                   const Constructor &constructor)
{
    Vector<const char *> vArgumentNames;
    
    X_U32 count = constructor.GetSignature().GetArgumentCount();

    for (X_U32 i = 0; i < count; i++) {
        vArgumentNames[i] = constructor.GetArgumentName(i);
    }
    
    PrintConstructorSignatureSuffix(out, constructor.GetSignature(), 
                                    &vArgumentNames);
    
    count = constructor.GetSignature().GetThrowCount();

    if (count > 0) {
        out.Write(" throw(");
        // For some reason gccxml outputs these in the opposite order from the
        // source code
        for (X_U32 i = count; i > 0; i--) {
            PrintTypePrefix(out, constructor.GetSignature().GetThrow(i - 1));
            if (i > 1) {
                out.Write(", ");
            }
        }
        out.Write(")");
    }

    out.WriteLine(";");
}

static void PrintMethod(OutputIndentStream &out, const Method &method)
{
    if (method.IsStatic()) {
        out.Write("static ");
    }

    if (method.IsVirtual()) {
        out.Write("virtual ");
    }

    Vector<const char *> vArgumentNames;
    
    X_U32 count = method.GetSignature().GetArgumentCount();

    for (X_U32 i = 0; i < count; i++) {
        vArgumentNames[i] = method.GetArgumentName(i);
    }

    // Print the return type
    PrintTypePrefix(out, method.GetSignature().GetReturnType());

    out.WriteF(" %s", method.GetName());

    PrintConstructorSignatureSuffix(out, method.GetSignature(), 
                                    &vArgumentNames);

    count = method.GetSignature().GetThrowCount();

    if (count > 0) {
        out.Write(" throw(");
        // For some reason gccxml outputs these in the opposite order from the
        // source code
        for (X_U32 i = count; i > 0; i--) {
            PrintTypePrefix(out, method.GetSignature().GetThrow(i - 1));
            if (i > 1) {
                out.Write(", ");
            }
        }
        out.Write(")");
    }

    if (method.IsConst()) {
        out.Write(" const");
    }

    if (method.IsPureVirtual()) {
        out.Write(" = 0");
    }

    out.WriteLine(";");
}

static void PrintEnumeration(OutputIndentStream &out,
                             const X_Enumeration &enumeration)
{
    out.WriteLineF("enum %s {", enumeration.GetName());
    out.indent();
    X_U32 count = enumeration.GetValueCount();
    for (X_U32 i = 0; i < count; i++) {
        const EnumerationValue &value = enumeration.GetValue(i);
        out.WriteF("%s = %ld", value.GetName(), value.GetValue());
        if (i < (count - 1)) {
            out.Write(",");
        }
        out.WriteLine();
    }
    out.deindent();
    out.WriteLine("};");
}

static void PrintValue(OutputIndentStream &out, const Type &type,
                       const Value &value)
{
    switch (type.GetBaseType()) {
    case Type::BaseType_Bool:
        out.Write(value.bool_v ? "true" : "false");
        break;
    case Type::BaseType_Char:
    case Type::BaseType_Unsigned_Char:
        out.WriteF("'%c'", value.char_v);
        break;
    case Type::BaseType_WChar:
    case Type::BaseType_Short:
        out.WriteF("%d", value.short_v);
        break;
    case Type::BaseType_Unsigned_Short:
        out.WriteF("%u", value.unsigned_short_v);
        break;
    case Type::BaseType_Int:
        out.WriteF("%ld", value.int_v);
        break;
    case Type::BaseType_Unsigned_Int:
        out.WriteF("%lu", value.unsigned_int_v);
        break;
    case Type::BaseType_Long:
        out.WriteF("%ld", value.long_v);
        break;
    case Type::BaseType_Unsigned_Long:
        out.WriteF("%lu", value.unsigned_long_v);
        break;
    case Type::BaseType_Long_Long:
        out.WriteF("%lld", value.long_long_v);
        break;
    case Type::BaseType_Unsigned_Long_Long:
        out.WriteF("%llu", value.unsigned_long_long_v);
        break;
    case Type::BaseType_Float:
        out.WriteF("%f", value.float_v);
        break;
    case Type::BaseType_Double:
        out.WriteF("%g", value.double_v);
        break;
    case Type::BaseType_Long_Double:
        out.WriteF("%lg", value.long_double_v);
    case Type::BaseType_Enumeration:
        out.WriteF("%d", value.int_v);
        break;
    case Type::BaseType_Function:
        {
            FunctionExtra &fe = FunctionExtra::Get
                (((const TypeFunction &) type).GetSignature());
            
            out.WriteF("function_%lu", fe.GetNumber());
        }
        break;
    // I don't think these makes sense ...
    case Type::BaseType_Bitfield:
    case Type::BaseType_Void:
    default: // BaseType_Structure: 
        out.Write("XXX-UNIMPLEMENTED-XXX");
        break;
    }
}

static void PrintTypePrefix(OutputIndentStream &out, const Type &type)
{
    String typestr;

    if (type.IsConst()) {
        typestr = "const ";
    }

    if (type.IsVolatile()) {
        typestr = "volatile ";
    }

    switch (type.GetBaseType()) {
    case Type::BaseType_Void:
        typestr += "void";
        break;
    case Type::BaseType_Bool:
        typestr += "bool";
        break;
    case Type::BaseType_Char:
        typestr += "char";
        break;
    case Type::BaseType_Unsigned_Char:
        typestr += "unsigned char";
        break;
    case Type::BaseType_WChar:
        typestr += "wchar_t";
        break;
    case Type::BaseType_Short:
        typestr += "short";
        break;
    case Type::BaseType_Unsigned_Short:
        typestr += "unsigned short";
        break;
    case Type::BaseType_Int:
        typestr += "int";
        break;
    case Type::BaseType_Unsigned_Int:
        typestr += "unsigned int";
        break;
    case Type::BaseType_Long:
        typestr += "long";
        break;
    case Type::BaseType_Unsigned_Long:
        typestr += "unsigned long";
        break;
    case Type::BaseType_Long_Long:
        typestr += "long long";
        break;
    case Type::BaseType_Unsigned_Long_Long:
        typestr += "unsigned long long";
        break;
    case Type::BaseType_Float:
        typestr += "float";
        break;
    case Type::BaseType_Double:
        typestr += "double";
        break;
    case Type::BaseType_Long_Double:
        typestr += "long double";
        break;
    case Type::BaseType_Bitfield:
        typestr += "int";
        break;
    case Type::BaseType_Enumeration:
        typestr += ((const TypeEnumeration &) type).GetEnumeration().GetName();
        break;
    case Type::BaseType_Function:
        {
            FunctionExtra &fe = FunctionExtra::Get
                (((const TypeFunction &) type).GetSignature());
            
            typestr = String::Format("function_%lu", fe.GetNumber());
        }
        break;
    default: { // BaseType_Structure: 
        const Structure &structure = 
            ((const TypeStructure &) type).GetStructure();
        
        if (structure.IsAnonymous()) {
            // Write the type out on the spot
            PrintStructure(out, (const Struct &) structure);
        }
        else {
            typestr += structure.GetName();
        }
    }
        break;
    }

    X_U32 count = type.GetPointerCount();

    if (count > 0) {
        typestr += " ";

        String p;

        for (X_U32 i = 0; i < count; i++) {
            const Pointer &pointer = type.GetPointer(i);

            if (pointer.IsConst()) {
                if (pointer.IsVolatile()) {
                    p -= " const volatile ";
                }
                else {
                    p -= " const ";
                }
            }
            else if (pointer.IsVolatile()) {
                p -= " volatile ";
            }

            p -= "*";
        }

        typestr += p;
    }
    
    if (type.IsReference()) {
        typestr += " &";
    }

    out.WriteF("%S", (const unichar *) typestr);
}

static void PrintTypeSuffix(OutputIndentStream &out, const Type &type)
{
    String typestr;

    if (type.GetBaseType() == Type::BaseType_Bitfield) {
        typestr += String::Format(":%lu", type.GetBitfieldBitCount());
    }
    else {
        X_U32 count = type.GetArrayCount();

        for (X_U32 i = 0; i < count; i++) {
            const Array &array = type.GetArray(i);
            
            if (array.IsUnbounded()) {
                typestr += "[]";
            }
            else {
                typestr += String::Format("[%lu]", array.GetElementCount());
            }
        }
    }

    out.Write(typestr);
}

static void PrintConstructorSignatureSuffix(OutputIndentStream &out, 
                                            const ConstructorSignature &sig,
                                            Vector<const char *> *pArgNames)
{
    out.Write("(");

    X_U32 count = sig.GetArgumentCount();

    for (X_U32 i = 0; i < count; i++) {
        const Argument &arg = sig.GetArgument(i);
        PrintTypePrefix(out, arg.GetType());
        if (pArgNames) {
            out.WriteF(" %s", pArgNames->Get(i));
        }
        if (arg.HasDefault()) {
            out.Write(" = (");
            PrintTypePrefix(out, arg.GetType());
            out.Write(") ");
            PrintValue(out, arg.GetType(), arg.GetDefault());
        }
        if (i < (count - 1)) {
            out.Write(", ");
        }
    }

    out.Write(")");
}

void XrttiToCpp(OutputStream &os)
{
    X_U32 count = GetContextCount();

    // First create a name-sorted list of contexts
    Vector<const Context *> vNameSortedContexts;

    for (X_U32 i = 0; i < count; i++) {
        const Context *pContext = GetContext(i);
        SortIntoByName(*pContext, vNameSortedContexts);
    }

    // Now create a hierarchy of contexts
    for (X_U32 i = 0; i < count; i++) {
        AddToHierarchy(*(vNameSortedContexts[i]));
    }

    // Now sort the hierarchy - starting with the root namespace, sort the
    // children of each context successively
    const Context *pRoot = LookupContext("::");
    if (!pRoot) {
        // If there is no root, then no contexts have been registered and
        // there is nothing to do
        return;
    }
    ContextExtra &rootExtra = ContextExtra::Get(*pRoot);
    rootExtra.SortChildren();

    // Collect the child enumerations of each context
    CollectEnumerations(*pRoot);
    
    // Use an output indent stream
    OutputIndentStream out(os);
    
    // First print out any function typedefs
    PrintFunctionTypedefs(out, *pRoot);

    // Now print the hierarchy, starting with children of the root node
    PrintChildrenContexts(out, *pRoot);
}
