/*****************************************************************************\
 *                                                                           *
 * Configuration.cpp                                                         *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include <Essential/ProgramVariables.h>
#include <Essential/Log.h>
#include <Essential/Util.h>
#include "../../inc/private/Configuration.h"

using namespace Essential;


static void UsageExit()
{
    Log::Warning
        ("Usage: xrttigen [-e <exclude_spec>]...\n"
         "                [-g <GENERATED_NAME>]\n"
         "                [-h <header_file>]\n"
         "                [-i <include_spec>]...\n"
         "                [-n]\n"
         "                [-o]\n"
         "                [-x]\n\n"
         "  -e:   Gives a specification of a class or a set of classes to\n"
         "        exclude from generation.  Includes and excludes are\n"
         "        evaluated in the order that they appear on the command\n"
         "        line, with later includes and excludes taking precedence.\n"
         "  -g:   Gives the name of the generated code; this defines the\n"
         "        name that the structures thus generated will have, and\n"
         "        can be used in a call to Xrtti::Register().  If this\n"
         "        parameter is not present, automatic registration of the\n"
         "        generated code will happen (no need to call\n"
         "        Xrtti::Register with the generated code).\n"
         "  -h:   Adds a header file to the set of header files to include\n"
         "        in the generated output.  This is typically the same set\n"
         "        of header files used in the generation of the input\n"
         "        gccxml output.\n"
         "  -i:   Gives a specification of a class or a set of classes to\n"
         "        include in generation.  Includes and excludes are\n"
         "        evaluated in the order that they appear on the command\n"
         "        line, with later includes and excludes taking precedence.\n"
         "  -n:   If present disables C++ RTTI support in the generated\n"
         "        Xrtti structures; this flag would be specified if your\n"
         "        C++ code is being built without C++ rtti support.\n"
         "  -o:   Names the output file to write the generated source to.  A\n"
         "        value of dash (-) indicates stdout.  Default is stdout.\n"
         "  -x:   Names the input file to read the gccxml source from.  A\n"
         "        value of dash (-) indicates stdin.  Default is stdin.\n\n"
         "  Include/Exclude Specifications are of the following form:\n\n"
         "  'name', or 'struct:name' : Identifies a specific class by name\n"
         "  'subof:name' : Identifies all classes which are subclasses of\n"
         "                 the named class\n"
         "  'has:name' : Identifies all classes which have a field with the\n"
         "               given name\n\n"
         "  All names may use simple wildcarding, in which either the\n"
         "  first or last character of the name is an asterisk\n\n");

    Util::Terminate();
}

static inline bool IsIdentifierBeginning(const char &uc)
{
    return (((uc >= 'a') && (uc <= 'z')) || ((uc >= 'A') && (uc <= 'Z')) ||
            (uc == '_'));
}

static inline bool IsIdentifierPostBeginning(const char &uc)
{
    return (IsIdentifierBeginning(uc) || ((uc >= '0') && (uc <= '9')));
}

Configuration::Configuration(int argc, char **argv)
{
    ProgramVariables vars(argc, argv);

    if (vars.GetBool("-?", "HELP") || vars.GetBool("--help", "HELP")) {
        UsageExit();
    }

    // -n
    disableRttiM = vars.GetBool("-n", "NO_RTTI");

    // -g
    generatedNameM = vars.GetString("-g", "GENERATED", "");
    
    // -h
    while (vars.HasVariable("-h", "HEADER")) {
        String header = vars.GetString("-h", "HEADER", "");
        vHeadersM.Append(header);
    }
    
    // -x (defaults to stdin)
    const char *x = vars.GetString("-x", "XML_INPUT", NULL);
    if ((x == NULL) || (x == "-")) {
        pInFileM = NULL;
        pInM = &(File::Stdin);
    }
    else {
        pInFileM = new File();
        if (!pInFileM->Open(x, File::OpenTypeRead)) {
            Log::ErrorF("Unable to open input file %s\n", x);
            UsageExit();
        }
        pInM = pInFileM;
    }
    
    // -o (defaults to stdout)
    const char *o = vars.GetString("-o", "OUTPUT", NULL);
    if ((o == NULL) || String::Equal(o, "-")) {
        pOutFileM = NULL;
        pOutM = &(File::Stdout);
    }
    else {
        pOutFileM = new File();
        if (!pOutFileM->Open(o, File::OpenTypeWriteTruncate)) {
            Log::ErrorF("Unable to open output file %s\n", o);
            UsageExit();
        }
        pOutM = pOutFileM;
    }

    // Get as of yet unused variables
    argc = vars.GetUnusedVariablesCount();
    argv = vars.GetUnusedVariables();

    // Skip program name
    argv++, argc--;

    // -i (or just +), -e (or just -)
    while (argc) {
        if (String::Equal(*argv, "-i")) {
            argc--, argv++;
            if (argc == 0) {
                Log::ErrorF("Parameter -i is missing argument\n");
                UsageExit();
            }
            ref<Clude> pClude = new Clude(Clude::TypeIn, *argv);
            vCludesM.Append(pClude);
        }
        else if (String::Equal(*argv, "-e")) {
            argc--, argv++;
            if (argc == 0) {
                Log::ErrorF("Parameter -i is missing argument\n");
                UsageExit();
            }
            ref<Clude> pClude = new Clude(Clude::TypeEx, *argv);
            vCludesM.Append(pClude);
        }
        else if (*argv[0] == '+') {
            ref<Clude> pClude = new Clude(Clude::TypeIn, &((*argv)[1]));
            if (pClude->name == "") {
                Log::ErrorF("Invalid parameter: %s\n", *argv);
                UsageExit();
            }
            vCludesM.Append(pClude);
        }
        else if (*argv[0] == '-') {
            ref<Clude> pClude = new Clude(Clude::TypeEx, &((*argv)[1]));
            if (pClude->name == "") {
                Log::ErrorF("Invalid parameter: %s\n", *argv);
                UsageExit();
            }
            vCludesM.Append(pClude);
        }
        else {
            Log::ErrorF("Unknown parameter (expected -i/-e): %s\n", *argv);
            UsageExit();
        }
        argv++, argc--;
    }
}

Configuration::~Configuration()
{
    delete pInFileM;
    delete pOutFileM;
}

bool Configuration::ShouldInclude(const String &name) const
{
    // Walk the list of "cludes" backwards, since each later one takes
    // precedence over prior ones
    U32 cludeCount = vCludesM.Count();

    for (U32 i = cludeCount; i > 0; i--) {
        ref<Clude> pClude = vCludesM[i - 1];

        // See if it's in there
        if (!pClude->Matches(name)) {
            // Nope, go to the next clude
            continue;
        }

        // OK, it matches - what does that mean?
        return (pClude->type == Clude::TypeIn) ? true : false;
    }

    // Default is to include everything
    return true;
}

Configuration::Clude::Clude(Clude::Type theType, char *theName)
{
    type = theType;

    String rawName = theName;

    if (rawName.StartsWith("subof:")) {
        match = MatchSubOf;
        name = String(&(theName[String::Length("subof:")]));
    }
    else if (rawName.StartsWith("has:")) {
        match = MatchHas;
        name = String(&(theName[String::Length("has:")]));
    }
    else {
        match = MatchExact;
    }

    if (rawName.StartsWith("*")) {
        wildcardStartM = true;
        wildcardEndM = false;
        name = String(&(theName[1]));
    }
    else if (rawName.EndsWith("*")) {
        wildcardStartM = false;
        wildcardEndM = true;
        name = String(theName, 0, String::Length(theName) - 1);
    }
    else {
        wildcardStartM = wildcardEndM = false;
        name = rawName;
    }
}

bool Configuration::Clude::Matches(const String &otherName)
{
    if (wildcardStartM) {
        return otherName.EndsWith(name);
    }
    else if (wildcardEndM) {
        return otherName.StartsWith(name);
    }
    else {
        return otherName.Equals(name);
    }
}
