/***************************************************************************** \
 *                                                                           *
 * Generator.cpp                                                             *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/

#include <Essential/Log.h>
#include "../../inc/private/Generator.h"


U32 Generator::Field::fieldNumberG = 0;
U32 Generator::Method::methodNumberG = 0;

// ---------------------------------------------------------------------------
// Generator::Generator(const Configuration &config, const Parser &parser)
// ---------------------------------------------------------------------------
Generator::Generator(const Configuration &config, const Parser &parser)
    : configM(config), parserM(parser), outM(config.GetOutputStream()), 
      isTrueArrayWrittenM(false), isFalseArrayWrittenM(false)
{
}


// ---------------------------------------------------------------------------
// Generator::Generate()
// ---------------------------------------------------------------------------
void Generator::Generate()
{
    // Iterate over all of the elements in the Parser
    Parser::ElementEnumeration elements(parserM);

    // For each Class element, determine if it is to be included in the
    // output.  If so, get it.  This populates the Generator object with
    // all of the classes/types/etc that will be generated.
    while (elements.HasElements()) {
        ref<Parser::Element> pElement = elements.GetNextElement();
        String elementName = pElement->GetElementName();

        // Skip it if it's not a Class, Struct, or Union
        if ((elementName != "Class") && (elementName != "Struct") &&
            (elementName != "Union")) {
            continue;
        }

        // Check the id to make sure it's OK
        String id = pElement->GetAttributeValue("id");
        if (id == "") {
            this->Error("%S element at line %lu is missing an id "
                        "attribute", (const unichar *) elementName, 
                        pElement->GetLineNumber());
        }
        
        // Skip it if it should not be included
        String fullName = this->GetFullNameOf(id);
        if (!configM.ShouldInclude(fullName)) {
            continue;
        }

        // Skip it if it's a gccxml-generated pseudo class.  We detect
        // these classes because they are defined in the File "<internal>"
        ref<Parser::Element> pFileElement = 
            this->LookupParserElement(pElement->GetAttributeValue("file"));
        if ((pFileElement != NULL) &&
            (pFileElement->GetAttributeValue("name") == "<internal>")) {
            continue;
        }

        // Also skip these very weird gccxml-generated classes
        if (fullName.StartsWith("__vmi_class_type_info_pseudo")) {
            continue;
        }

        // OK, it's a Class/Struct/Union, we'll include it

        // Get it, which populates the hashtables and vectors with it and
        // everything it references
        (void) GetContext(id);
    }

    // Emit the header into the source
    this->EmitHeader();

    // Emit the typedefs into the source
    this->EmitTypedefs();

    // Emit the generated into the source
    this->EmitGenerated();
}


// ---------------------------------------------------------------------------
// Generator::Error(const String &format, ...)
// ---------------------------------------------------------------------------
void Generator::Error(const String &format, ...)
{
    va_list args;
        
    va_start(args, format);

    Log::ErrorV(format, args);
    
    va_end(args);

    Util::Terminate();
}


// ---------------------------------------------------------------------------
// Generator::GetContext(const String &id)
// ---------------------------------------------------------------------------
ref<Generator::Context> Generator::GetContext(const String &id)
{
    ref<Context> pContext = htContextsM.Get(id);

    if (pContext != NULL) {
        return pContext;
    }

    // A Context is either a Namespace or Class; figure out which
    ref<Parser::Element> pElement = this->LookupParserElement(id);

    if (pElement == NULL) {
        this->Error("Bad id: %S\n", (const unichar *) id);
    }

    String elementName = pElement->GetElementName();
    
    if (elementName == "Namespace") {
        ref<Namespace> pNamespace = new Namespace(*this, id);
        pNamespace->SetNumber(htContextsM.Count());
        htContextsM.Put(id, (Context *) pNamespace);
        pNamespace->Initialize();
        return (Context *) pNamespace;
    }
    else if ((elementName == "Class") ||
             (elementName == "Struct") ||
             (elementName == "Union")) {
        ref<Class> pClass = new Class(*this, id);
        pClass->SetNumber(htContextsM.Count());
        htContextsM.Put(id, (Context *) pClass);
        pClass->Initialize();
        return (Context *) pClass;
    }
    else {
        this->Error("Expected Namespace, Class, Struct, or Union for " 
                    "id %S; got %S\n", (const unichar *) id, 
                    (const unichar *) elementName);
        // This will not be reached
        return NULL;
    }
}


// ---------------------------------------------------------------------------
// Generator::GetEnumeration(const String &id)
// ---------------------------------------------------------------------------
ref<Generator::Enumeration> Generator::GetEnumeration(const String &id)
{
    ref<Enumeration> pEnumeration = htEnumerationsM.Get(id);

    if (pEnumeration != NULL) {
        return pEnumeration;
    }

    // Make a new one and put it in the hashtable and vector
    pEnumeration = new Enumeration(*this, id);
    pEnumeration->SetNumber(htEnumerationsM.Count());
    htEnumerationsM.Put(id, pEnumeration);
    pEnumeration->Initialize();
    return pEnumeration;
}


// ---------------------------------------------------------------------------
// Generator::GetType(const String &id)
// ---------------------------------------------------------------------------
ref<Generator::Type> Generator::GetType(const String &id, const String &bits)
{
    // Hash types by id and bits
    String key = id + "_" + bits;

    ref<Type> pType = htTypesM.Get(key);

    if (pType != NULL) {
        return pType;
    }

    // Make a new one and put it in the hashtable and vector
    pType = new Type(*this, id, bits);
    pType->SetNumber(htTypesM.Count());
    htTypesM.Put(key, pType);
    pType->Initialize();
    return pType;
}


// ---------------------------------------------------------------------------
// Generator::EmitHeader()
// ---------------------------------------------------------------------------
void Generator::EmitHeader()
{
    // Emit the disclaimer
    outM.WriteA
        ("/****************************************************"
         "*************************\\\n"
         " *                                                         "
         "                  *\n"
         " *    AUTOGENERATED FILE - DO NOT EDIT - "
         "AUTOGENERATED FILE - DO NOT EDIT    *\n"
         " *                                                         "
         "                  *\n"
         " *      This file was generated by the "
         "xrttigen program.                     *\n"
         " *      Do not edit this file directly.  Edit the source "
         "file(s) instead     *\n"
         " *      and re-generate this file.                         "
         "                  *\n"
         " *                                                          "
         "                 *\n"
         " *    AUTOGENERATED FILE - DO NOT EDIT - "
         "AUTOGENERATED FILE - DO NOT EDIT    *\n"
         " *                                                          "
         "                 *\n"
         "\\************************************************************"
         "*****************/\n\n");

    // Emit the "includes"
    outM.WriteA("#include <Xrtti/XrttiPrivate.h>\n");
    U32 count = configM.GetHeaderCount();
    for (U32 i = 0; i < count; i++) {
        String header = configM.GetHeader(i);
        if (header.StartsWith("<")) {
            outM.WriteFA("#include %S\n", (const unichar *) header);
        }
        else {
            outM.WriteFA("#include \"%S\"\n", (const unichar *) header);
        }
    }

    outM.WriteA("\n");
}


// ---------------------------------------------------------------------------
// Generator::EmitTypedefs()
// ---------------------------------------------------------------------------
void Generator::EmitTypedefs()
{
    // Emit the typedefs for each class; this is necessary because we need a
    // name that we can use in the XrttiPrivate_offsetof() macro for each
    // type, and templated types cannot be used there because of the commas
    RefHashtable__ValueEnumeration<String, Context> contexts(htContextsM);

    bool atLeastOne = false;

    while (contexts.HasElements()) {
        ref<Context> pContext = contexts.GetNextElement();
        if (pContext->GetContextType() != Context::ContextTypeClass) {
            continue;
        }
        ref<Class> pClass = (Class *) (Context *) pContext;
        if (pClass->IsAnonymous() || pClass->IsIncomplete()) {
            continue;
        }
        pClass->EmitTypedef();
        atLeastOne = true;
    }

    if (atLeastOne) {
        outM.WriteA("\n");
    }
}


// ---------------------------------------------------------------------------
// Generator::EmitGenerated()
// ---------------------------------------------------------------------------
void Generator::EmitGenerated()
{
    // Emit the contexts and their array
    U32 count = htContextsM.Count();
    if (count > 0) {
        RefHashtable__ValueEnumeration<String, Context> contexts(htContextsM);

        while (contexts.HasElements()) {
            contexts.GetNextElement()->Emit();
        }

        outM.WriteFA("static const XP_Context * const "
                       "XrttiPrivate_Contexts[%lu] =\n{\n", count);

        for (U32 i = 0; i < count; i++) {
            outM.WriteFA("    & XrttiPrivate_Context_%lu", i);
            if (i < (count - 1)) {
                outM.WriteA(",\n");
            }
            else {
                outM.WriteA("\n");
            }
        }

        outM.WriteA("};\n\n");
    }

    // Emit the enumerations array
    count = htEnumerationsM.Count();
    if (count > 0) {
        outM.WriteFA("static const XP_Enumeration * const "
                       "XrttiPrivate_Enumerations[%lu] =\n{\n", count);

        for (U32 i = 0; i < count; i++) {
            outM.WriteFA("    & XrttiPrivate_Enumeration_%lu", i);
            if (i < (count - 1)) {
                outM.WriteA(",\n");
            }
            else {
                outM.WriteA("\n");
            }
        }

        outM.WriteA("};\n\n");
    }

    // Emit the types array
    count = htTypesM.Count();
    if (count > 0) {
        outM.WriteFA("static const XP_Type * const "
                       "XrttiPrivate_Types[%lu] =\n{\n", count);

        for (U32 i = 0; i < count; i++) {
            outM.WriteFA("    & XrttiPrivate_Type_%lu", i);
            if (i < (count - 1)) {
                outM.WriteA(",\n");
            }
            else {
                outM.WriteA("\n");
            }
        }

        outM.WriteA("};\n\n");
    }

    String generatedName = configM.GetGeneratedName();
    if (generatedName == "") {
        // No generated name means auto-register, using a static
        outM.WriteA("static const X_Generated generated =\n{\n");
    }
    else {
        outM.WriteFA("extern const X_Generated %S =\n{\n", 
                     (const unichar *) generatedName);
    }

    outM.WriteFA("    /* contexts_count */ %lu,\n", htContextsM.Count());
    if (htContextsM.Count()) {
        outM.WriteA("    /* contexts */ XrttiPrivate_Contexts,\n");
    }
    else {
        outM.WriteA("    /* contexts */ 0,\n");
    }

    outM.WriteFA("    /* enumerations_count */ %lu,\n", 
                   htEnumerationsM.Count());
    if (htEnumerationsM.Count()) {
        outM.WriteA("    /* enumerations */ XrttiPrivate_Enumerations,\n");
    }
    else {
        outM.WriteA("    /* enumerations */ 0,\n");
    }

    outM.WriteFA("    /* types_count */ %lu,\n", htTypesM.Count());
    if (htTypesM.Count()) {
        outM.WriteA("    /* types */ XrttiPrivate_Types\n");
    }
    else {
        outM.WriteA("    /* types */ 0\n");
    }

    outM.WriteA("};\n");

    // Now write the AutoRegister if necessary
    if (generatedName == "") {
        outM.WriteFA("\nstatic X_AutoRegister autoRegister(generated);\n");
    }
}


// ---------------------------------------------------------------------------
// Generator::EmitSpecialArray(bool trueArray)
// ---------------------------------------------------------------------------
void Generator::EmitSpecialArray(bool trueArray)
{
    if (trueArray) {
        if (!isTrueArrayWrittenM) {
            outM.WriteA("static const bool XrttiPrivate_Xrtti_true[] ="
                          "\n{\n    true\n};\n\n");
            isTrueArrayWrittenM = true;
        }
    }
    else {
        if (!isFalseArrayWrittenM) {
            outM.WriteA("static const bool XrttiPrivate_Xrtti_false[] ="
                          "\n{\n    false\n};\n\n");
            isFalseArrayWrittenM = true;
        }
    }
}


// ---------------------------------------------------------------------------
// Generator::EmitSpecialArrayReference(bool trueArray, const char *&name)
// ---------------------------------------------------------------------------
void Generator::EmitSpecialArrayReference(bool trueArray, const char *name)
{
    if (trueArray) {
        outM.WriteFA("    /* %s */ XrttiPrivate_Xrtti_true,\n", name);
    }
    else {
        outM.WriteFA("    /* %s */ XrttiPrivate_Xrtti_false,\n", name);
    }
}


// ---------------------------------------------------------------------------
// Generator::split(const char *pString, RefVector<RefString> &v)
// ---------------------------------------------------------------------------
/* static */
void Generator::split(const unichar *pString, RefVector<RefString> &v)
{
    const unichar *p = pString;
    while (true) {
        // Skip whitespace to new member
        while (*p && (*p == ' ')) {
            p++;
        }
        if (!*p) {
            break;
        }
        // This is the start of the member
        pString = p;
        // Skip to whitespace
        while (*p && (*p != ' ')) {
            p++;
        }
        // Now add the id
        ref<RefString> pValue = new RefString(pString, 0, p - pString);
        v.Append(pValue);
    }
}


// ---------------------------------------------------------------------------
// Generator::GetFullNameOf(const String &elementId)
// ---------------------------------------------------------------------------
String Generator::GetFullNameOf(const String &elementId)
{
    ref<Parser::Element> pElement = this->LookupParserElement(elementId);

    if (pElement == NULL) {
        return "";
    }

    String name = pElement->GetAttributeValue("name");

    String parentName = this->GetFullNameOf
        (pElement->GetAttributeValue("context"));

    if ((parentName == "") || (parentName == "::")) {
        return name;
    }

    return parentName + "::" + name;
}


// ---------------------------------------------------------------------------
// Generator::Namespace::Emit()
// ---------------------------------------------------------------------------
void Generator::Namespace::Emit()
{
    // Don't emit anything if it's already been emitted
    if (!this->ShouldEmit()) {
        return;
    }

    OutputStream &out = generatorM.GetOutputStream();
        
    out.WriteFA("static const XP_Context "
                  "XrttiPrivate_Context_%lu =\n{\n", this->GetNumber());

    // type
    out.WriteA("    /* type */ ::Xrtti::Context::Type_Namespace,\n");

    // name
    out.WriteFA("    /* name */ \"%S\",\n", 
                  (const unichar *) this->GetName());

    // full
    out.WriteFA("    /* full_name */ \"%S\",\n", 
                  (const unichar *) this->GetFullName());

    // parent
    ref<Context> pParent = this->GetContext();
    if (pParent) {
        out.WriteFA("    /* parent */ %lu\n", pParent->GetNumber());
    }
    else {
        out.WriteA("    /* parent */ 0xFFFFFFFF\n");
    }

    out.WriteA("};\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Enumeration::Initialize()
// ---------------------------------------------------------------------------
void Generator::Enumeration::Initialize()
{
    pContextM = generatorM.GetContext(pElementM->GetAttributeValue("context"));

    U32 count = pElementM->GetSubElementCount();

    for (U32 i = 0; i < count; i++) {
        ref<Parser::Element> pSubElement = pElementM->GetSubElement(i);
        
        if (pSubElement->GetElementName() != "EnumValue") {
            continue;
        }

        String name = pSubElement->GetAttributeValue("name");
        vNamesM.Append(name);
        S32 value = (S32) (pSubElement->GetAttributeValue("init").ToS64());
        vValuesM.Append(value);
    }
}


// ---------------------------------------------------------------------------
// Generator::Enumeration::Emit()
// ---------------------------------------------------------------------------
void Generator::Enumeration::Emit()
{
    if (!this->ShouldEmit()) {
        return;
    }

    OutputStream &out = generatorM.GetOutputStream();

    // Emit the value names array
    U32 count = vNamesM.Count();
    out.WriteFA("static const char * const XrttiPrivate_Enumeration_%lu_"
                  "value_names[%lu] =\n{\n", this->GetNumber(), count);
    
    for (U32 i = 0; i < count; i++) {
        out.WriteFA("    \"%S\"", (const unichar *) vNamesM[i]);
        if (i < (count - 1)) {
            out.WriteA(",\n");
        }
    }

    out.WriteA("\n};\n\n");

    // Emit the value values array
    out.WriteFA("static const X_S32 XrttiPrivate_Enumeration_%lu_"
                  "value_values[%lu] =\n{\n", this->GetNumber(), count);
    
    for (U32 i = 0; i < count; i++) {
        out.WriteFA("    %ld", (const unichar *) vValuesM[i]);
        if (i < (count - 1)) {
            out.WriteA(",\n");
        }
    }

    out.WriteA("\n};\n\n");

    // Emit the Enumeration itself
    out.WriteFA("static const XP_Enumeration XrttiPrivate_"
                  "Enumeration_%lu =\n{\n", this->GetNumber());

    // access_type
    out.WriteFA("    /* access_type */ %s,\n", this->GetAccessType());

    // context
    if (pContextM == NULL) {
        out.WriteA("    /* context */ 0xFFFFFFFF,\n");
    }
    else {
        out.WriteFA("    /* context */ %lu,\n", pContextM->GetNumber());
    }

    // name
    out.WriteFA("    /* name */ \"%S\",\n", 
                  (const unichar *) this->GetName());

    // values_count
    out.WriteFA("    /* values_count */ %lu,\n", count);
    
    // value_names
    if (count > 0) {
        out.WriteFA("    /* value_names */ XrttiPrivate_Enumeration_%lu_"
                      "value_names,\n", this->GetNumber());
    }
    else {
        out.WriteA("    /* value_names */ 0,\n");
    }

    // value_values
    if (count > 0) {
        out.WriteFA("    /* value_values */ XrttiPrivate_Enumeration_%lu_"
                      "value_values\n", this->GetNumber());
    }
    else {
        out.WriteA("    /* value_values */ 0\n");
    }
    
    out.WriteFA("};\n\n");
}

// ---------------------------------------------------------------------------
// Generator::Type::Initialize()
// ---------------------------------------------------------------------------
void Generator::Type::Initialize()
{
    constref<Parser::Element> pElement = pElementM;
    
    // Now first we may have reference type
    if (pElement->GetElementName() == "ReferenceType") {
        // Set reference to true
        isReferenceM = true;
        // Get the type that is referenced
        pElement = this->GetCppTypeElement(pElement);
    }
    else {
        isReferenceM = false;
    }

    // Now loop through the remaining types
    Attributes attributes = { false, false };
    while (true) {
        String type = pElement->GetElementName();

        // CvQualifiedType simply adds qualifying attributes to the next
        // type specifier
        if (type == "CvQualifiedType") {
            attributes.isConst = (pElement->GetAttributeValue("const") == "1");
            attributes.isVolatile =
                (pElement->GetAttributeValue("volatile") == "1");
            // Ignore restrict, it is g++-specific
        }
        // PointerType adds to the pointer level vector, with the last
        // used attributes
        else if (type == "PointerType") {
            vPointerAttributesM.Append(attributes);
            attributes.isConst = attributes.isVolatile = false;
        }
        // ArrayType adds another array dimension, with its max
        else if (type == "ArrayType") {
            String max = pElement->GetAttributeValue("max");
            MaxIndex maxIndex;
            if ((max == "ffffffffffffffff") ||
                (max == "0xffffffffffffffff")) {
                // This indicates no practical maximumum
                maxIndex.isUnbounded = true;
                maxIndex.max = 0;
            }
            else {
                maxIndex.isUnbounded = false;
                maxIndex.max = max.ToU64();
            }

            vArrayDimensionMaxIndexesM.Append(maxIndex);
        }
        // Skip right over typedefs, to the actual type
        else if (type == "Typedef") {
        }
        // If bits is set, this is a bitfield
        else if (bitsM != "") {
            cppTypeM = CppTypeBitfield;
            break;
        }
        // Fundamental types finish the type definition
        else if (type == "FundamentalType") {
            String name = pElement->GetAttributeValue("name");

            if (name == "void") {
                cppTypeM = CppTypeVoid;
            }
            else if (name == "bool") {
                cppTypeM = CppTypeBool;
            }
            else if (name == "char") {
                cppTypeM = CppTypeChar;
            }
            else if (name == "signed char") {
                cppTypeM = CppTypeSignedChar;
            }
            else if (name == "unsigned char") {
                cppTypeM = CppTypeUnsignedChar;
            }
            else if (name == "wchar_t") {
                cppTypeM = CppTypeWChar;
            }
            else if (name == "short int") {
                cppTypeM = CppTypeShort;
            }
            else if (name == "short signed int") {
                cppTypeM = CppTypeSignedShort;
            }
            else if (name == "short unsigned int") {
                cppTypeM = CppTypeUnsignedShort;
            }
            else if (name == "int") {
                cppTypeM = CppTypeInt;
            }
            else if (name == "signed int") {
                cppTypeM = CppTypeSignedInt;
            }
            else if (name == "unsigned int") {
                cppTypeM = CppTypeUnsignedInt;
            }
            else if (name == "long int") {
                cppTypeM = CppTypeLong;
            }
            else if (name == "long signed int") {
                cppTypeM = CppTypeLong;
            }
            else if (name == "long unsigned int") {
                cppTypeM = CppTypeUnsignedLong;
            }
            else if (name == "long long int") {
                cppTypeM = CppTypeLongLong;
            }
            else if (name == "long long signed int") {
                cppTypeM = CppTypeSignedLongLong;
            }
            else if (name == "long long unsigned int") {
                cppTypeM = CppTypeUnsignedLongLong;
            }
            else if (name == "float") {
                cppTypeM = CppTypeFloat;
            }
            else if (name == "double") {
                cppTypeM = CppTypeDouble;
            }
            else if (name == "long double") {
                cppTypeM = CppTypeLongDouble;
            }
            else {
                generatorM.Error("Unexpected fundamental type: %S\n",
                                 (const unichar *) name);
            }
            break;
        }
        else if (type == "Enumeration") {
            cppTypeM = CppTypeEnumeration;
            String id = pElement->GetAttributeValue("id");
            pCppTypeEnumerationM = generatorM.GetEnumeration(id);
            if (pCppTypeEnumerationM == NULL) {
                generatorM.Error("Unknown %S type: %S\n", 
                                 (const unichar *) type,
                                 (const unichar *) id);
            }
            break;
        }
        else if (type == "FunctionType") {
            cppTypeM = CppTypeFunction;
            String id = pElement->GetAttributeValue("id");
            pCppTypeMethodM = new Method(generatorM, id, 
                                         Method::MethodTypeFunctionSignature);
            if (pCppTypeMethodM == NULL) {
                generatorM.Error("Unknown %S type: %S\n", 
                                 (const unichar *) type,
                                 (const unichar *) id);
            }
            break;
        }
        // Class types finish the type definition
        else if ((type == "Class") || (type == "Struct") || 
                 (type == "Union")) {
            cppTypeM = CppTypeClass;
            String id = pElement->GetAttributeValue("id");
            ref<Context> pContext = generatorM.GetContext(id);
            if ((pContext == NULL) || 
                (pContext->GetContextType() != Context::ContextTypeClass)) {
                generatorM.Error("Unknown %S type: %S\n", 
                                 (const unichar *) type,
                                 (const unichar *) id);
            }
            
            pCppTypeClassM = (Class *) (Context *) pContext;
            break;
        }
        else {
            generatorM.Error
                ("Unexpected type %S referenced by %S\n",
                 (const unichar *) type, 
                 (const unichar *) pElement->GetAttributeValue("id"));
        }
        
        // Next sub type
        pElement = this->GetCppTypeElement(pElement);
    }

    // Set attributes from last read CvQualifiedType
    attributesM = attributes;
}


// ---------------------------------------------------------------------------
// Generator::Type::Emit()
// ---------------------------------------------------------------------------
void Generator::Type::Emit()
{
    if (!this->ShouldEmit()) {
        return;
    }

    OutputStream &out = generatorM.GetOutputStream();
    
    // Ensure that the enumeration is written
    if (pCppTypeEnumerationM != NULL) {
        pCppTypeEnumerationM->Emit();
    }

    // Ensure that the method is written and a signature for it too
    if (pCppTypeMethodM != NULL) {
        pCppTypeMethodM->EmitSignature();
        pCppTypeMethodM->Emit(NULL);
    }

    // Emit the pointers const and volatile arrays
    U32 count = vPointerAttributesM.Count();

    if (count == 1) {
        generatorM.EmitSpecialArray(vPointerAttributesM[0].isConst);
        generatorM.EmitSpecialArray(vPointerAttributesM[0].isVolatile);
    }
    else if (count > 1) {
        out.WriteFA("static const bool XrttiPrivate_Type_%lu_pointers_"
                      "const[%lu] =\n{\n", this->GetNumber(), count);

        for (U32 i = 0; i < count; i++) {
            out.WriteFA("    %s", vPointerAttributesM[i].isConst ?
                          "true" : "false");

            if (i < (count - 1)) {
                out.WriteA(",\n");
            }
        }

        out.WriteA("\n};\n\n");

        out.WriteFA("static const bool XrttiPrivate_Type_%lu_pointers_"
                      "volatile[%lu] =\n{\n", this->GetNumber(), count);

        for (U32 i = 0; i < count; i++) {
            out.WriteFA("    %s", vPointerAttributesM[i].isVolatile ?
                          "true" : "false");

            if (i < (count - 1)) {
                out.WriteA(",\n");
            }
        }

        out.WriteA("\n};\n\n");
    }

    count = vArrayDimensionMaxIndexesM.Count();
    if (count == 1) {
        generatorM.EmitSpecialArray(vArrayDimensionMaxIndexesM[0].isUnbounded);
    }
    else if (count > 1) {
        out.WriteFA("static const bool XrttiPrivate_Type_%lu_arrays_"
                      "unbounded[%lu] =\n{\n", this->GetNumber(), count);

        for (U32 i = 0; i < count; i++) {
            out.WriteFA("    %s", vArrayDimensionMaxIndexesM[i].isUnbounded ?
                       "true" : "false");

            if (i < (count - 1)) {
                out.WriteA(",\n");
            }
        }
    
        out.WriteA("\n};\n\n");
    }

    if (count > 0) {
        out.WriteFA("static const X_U32 XrttiPrivate_Type_%lu_arrays_"
                    "element_counts[%lu] =\n{\n", this->GetNumber(), count);

        for (U32 i = 0; i < count; i++) {
            if (vArrayDimensionMaxIndexesM[i].isUnbounded) {
                out.WriteA("    0");
            }
            else {
                out.WriteFA("    %lu", vArrayDimensionMaxIndexesM[i].max + 1);
            }

            if (i < (count - 1)) {
                out.WriteA(",\n");
            }
        }
    
        out.WriteA("\n};\n\n");
    }

    // Emit the Type itself
    out.WriteFA("static const XP_Type XrttiPrivate_Type_%lu =\n{\n",
                  this->GetNumber());

    out.WriteFA("    /* type */ %s,\n", this->GetCppXrttiName());

    // enumeration (index into the array of enumerations)
    if (pCppTypeEnumerationM == NULL) {
        out.WriteA("    /* enumeration */ 0,\n");
    }
    else {
        out.WriteFA("    /* enumeration */ %lu,\n",
                      pCppTypeEnumerationM->GetNumber());
    }
    
    // method
    if (pCppTypeMethodM == NULL) {
        out.WriteA("    /* method */ 0,\n");
    }
    else {
        out.WriteFA("    /* method */ & XrttiPrivate_Method_%lu,\n",
                      pCppTypeMethodM->GetNumber());
    }

    // structure (index into the array of structures)
    if (pCppTypeClassM == NULL) {
        out.WriteA("    /* structure */ 0,\n");
    }
    else {
        out.WriteFA("    /* structure */ %lu,\n",
                      pCppTypeClassM->GetNumber());
    }

    // flags - is_const
    out.WriteFA("    /* is_const */ %s,\n",
                  attributesM.isConst ? "true" : "false");

    // flags - is_volatile
    out.WriteFA("    /* is_volatile */ %s,\n",
                  attributesM.isVolatile ? "true" : "false");

    // flags - is_reference
    out.WriteFA("    /* is_reference */ %s,\n",
                     isReferenceM ? "true" : "false");

    // bit count
    if (cppTypeM == CppTypeBitfield) {
        out.WriteFA("    /* bit_count */ %S,\n", (const unichar *) bitsM);
    }
    else {
        out.WriteA("    /* bit_count */ 0,\n");
    }

    // pointers_count
    count = vPointerAttributesM.Count();
    out.WriteFA("    /* pointers_count */ %lu,\n", count);

    // pointers_const
    // pointers_volatile
    if (count == 0) {
        out.WriteA("    /* pointers_const */ 0,\n");
        out.WriteA("    /* pointers_volatile */ 0,\n");
    }
    else if (count == 1) {
        generatorM.EmitSpecialArrayReference
            (vPointerAttributesM[0].isConst, "pointers_const");
        generatorM.EmitSpecialArrayReference
            (vPointerAttributesM[0].isVolatile, "pointers_volatile");
    }
    else {
        out.WriteFA("    /* pointers_const */ XrttiPrivate_Type_%lu_"
                      "pointers_const,\n", this->GetNumber());
        out.WriteFA("    /* pointers_volatile */ XrttiPrivate_Type_%lu_"
                      "pointers_volatile,\n", this->GetNumber());
    }

    // arrays_count
    count = vArrayDimensionMaxIndexesM.Count();
    out.WriteFA("    /* arrays_count */ %lu,\n", count);

    // arrays_unbounded
    if (count == 0) {
        out.WriteA("    /* arrays_unbounded */ 0,\n");
    }
    else if (count == 1) {
        generatorM.EmitSpecialArrayReference
            (vArrayDimensionMaxIndexesM[0].isUnbounded, "arrays_unbounded");
    }
    else {
        out.WriteFA("    /* arrays_unbounded */ XrttiPrivate_Type_%lu_"
                      "arrays_unbounded,\n", this->GetNumber());
    }

    // arrays_elemement_counts
    if (count == 0) {
        out.WriteA("    /* arrays_element_counts */ 0\n");
    }
    else {
        out.WriteFA("    /* arrays_element_counts */ XrttiPrivate_Type_"
                    "%lu_arrays_element_counts\n", this->GetNumber());
    }

    out.WriteA("};\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Type::IsAssignable()
// ---------------------------------------------------------------------------
bool Generator::Type::IsAssignable()
{
    // Not settable if const
    if (attributesM.isConst) {
        return false;
    }

    // If it's an anonymous class, it cannot be set
    if ((cppTypeM == CppTypeClass) && pCppTypeClassM->IsAnonymous()) {
        return false;
    }
    
    // If it's an array, it cannot be set
    if (this->GetArrayDimensionsCount() && !this->GetPointerDepth()) {
        return false;
    }
    
    // Otherwise, it can be set
    return true;
}


// ---------------------------------------------------------------------------
// Generator::Type::GetCppType(const Parser::Element *p)
// ---------------------------------------------------------------------------
constref<Parser::Element> Generator::Type::
GetCppTypeElement(const Parser::Element *p)
{
    String type = p->GetAttributeValue("type");

    ref<Parser::Element> pElement = generatorM.LookupParserElement(type);

    if (pElement == NULL) {
        generatorM.Error("Unknown sub type of type %S: %S, \n",
                         (const unichar *) p->GetAttributeValue("id"),
                         (const unichar *) type);
    }

    return pElement;
}


// --------------------------------------------------------------------------
// Generator::Type::GetCppFullType(bool includeConst)
// --------------------------------------------------------------------------
String Generator::Type::GetCppFullType(bool includeConst)
{
    String ret;

    // const modifier
    if (includeConst && attributesM.isConst) {
        ret = "const ";
    }

    // basic type
    ret += this->GetCppTypeName();
    // Use a pointer level for each pointer and for each array
    U32 pointerCount = (vPointerAttributesM.Count() + 
                        vArrayDimensionMaxIndexesM.Count());

    if (pointerCount) {
        ret += " ";

        for (U32 i = 0; i < pointerCount; i++) {
            ret += "*";
        }
    }
    
    return ret;
}


// ---------------------------------------------------------------------------
// Generator::Type::GetCppTypeXrttiName()
// ---------------------------------------------------------------------------
const char *Generator::Type::GetCppXrttiName()
{
    switch (cppTypeM) {
    case CppTypeVoid:
        return "X_Type::BaseType_Void";
    case CppTypeBool:
        return "X_Type::BaseType_Bool";
    case CppTypeChar:
    case CppTypeSignedChar:
        return "X_Type::BaseType_Char";
    case CppTypeUnsignedChar:
        return "X_Type::BaseType_Unsigned_Char";
    case CppTypeWChar:
        return "X_Type::BaseType_WChar";
    case CppTypeShort:
    case CppTypeSignedShort:
        return "X_Type::BaseType_Short";
    case CppTypeUnsignedShort:
        return "X_Type::BaseType_Unsigned_Short";
    case CppTypeInt:
    case CppTypeSignedInt:
        return "X_Type::BaseType_Int";
    case CppTypeUnsignedInt:
        return "X_Type::BaseType_Unsigned_Int";
    case CppTypeLong:
    case CppTypeSignedLong:
        return "X_Type::BaseType_Long";
    case CppTypeUnsignedLong:
        return "X_Type::BaseType_Unsigned_Long";
    case CppTypeLongLong:
    case CppTypeSignedLongLong:
        return "X_Type::BaseType_Long_Long";
    case CppTypeUnsignedLongLong:
        return "X_Type::BaseType_Unsigned_Long_Long";
    case CppTypeFloat:
        return "X_Type::BaseType_Float";
    case CppTypeDouble:
        return "X_Type::BaseType_Double";
    case CppTypeLongDouble:
        return "X_Type::BaseType_Long_Double";
    case CppTypeBitfield:
        return "X_Type::BaseType_Bitfield";
    case CppTypeEnumeration:
        return "X_Type::BaseType_Enumeration";
    case CppTypeFunction:
        return "X_Type::BaseType_Function";
    default: // case CppTypeClass
        return "X_Type::BaseType_Structure";
    }
}


// ---------------------------------------------------------------------------
// Generator::Type::GetCppTypeName()
// ---------------------------------------------------------------------------
String Generator::Type::GetCppTypeName()
{
    switch (cppTypeM) {
    case CppTypeVoid:
        return "void";
    case CppTypeBool:
        return "bool";
    case CppTypeChar:
        return "char";
    case CppTypeSignedChar:
        return "signed char";
    case CppTypeUnsignedChar:
        return "unsigned char";
    case CppTypeWChar:
        return "wchar_t";
    case CppTypeShort:
        return "short";
    case CppTypeSignedShort:
        return "signed short";
    case CppTypeUnsignedShort:
        return "unsigned short";
    case CppTypeBitfield:
    case CppTypeInt:
        return "int";
    case CppTypeSignedInt:
        return "signed int";
    case CppTypeUnsignedInt:
        return "unsigned int";
    case CppTypeLong:
        return "long";
    case CppTypeSignedLong:
        return "signed long";
    case CppTypeUnsignedLong:
        return "unsigned long";
    case CppTypeLongLong:
        return "long long";
    case CppTypeSignedLongLong:
        return "signed long long";
    case CppTypeUnsignedLongLong:
        return "unsigned long long";
    case CppTypeFloat:
        return "float";
    case CppTypeDouble:
        return "double";
    case CppTypeLongDouble:
        return "long double";
    case CppTypeEnumeration:
        return pCppTypeEnumerationM->GetFullName();
    case CppTypeFunction:
        return pCppTypeMethodM->GetSignature();
    default: // case CppTypeClass
        return pCppTypeClassM->GetFullName();
    }
}


// --------------------------------------------------------------------------
// Generator::Type::GetValueField()
// Gets the Value field which should be used for an argument of this Type
// --------------------------------------------------------------------------
String Generator::Type::GetValueField()
{
    if (this->GetPointerDepth() || this->GetArrayDimensionsCount()) {
        return String("pointer_v");
    }

    switch (cppTypeM) {
    case CppTypeBool:
        return String("bool_v");
    case CppTypeChar:
    case CppTypeSignedChar:
        return String("char_v");
    case CppTypeUnsignedChar:
        return String("unsigned_char_v");
    case CppTypeWChar:
        return String("wchar_v");
    case CppTypeShort:
    case CppTypeSignedShort:
        return String("short_v");
    case CppTypeUnsignedShort:
        return String("unsigned_short_v");
    case CppTypeInt:
    case CppTypeBitfield:
    case CppTypeEnumeration:
    case CppTypeSignedInt:
        return String("int_v");
    case CppTypeUnsignedInt:
        return String("unsigned_int_v");
    case CppTypeLong:
    case CppTypeSignedLong:
        return String("long_v");
    case CppTypeUnsignedLong:
        return String("unsigned_long_v");
    case CppTypeLongLong:
    case CppTypeSignedLongLong:
        return String("long_long_v");
    case CppTypeUnsignedLongLong:
        return String("unsigned_long_long_v");
    case CppTypeFloat:
        return String("float_v");
    case CppTypeDouble:
        return String("double_v");
    case CppTypeLongDouble:
        return String("long_double_v");
    default:
        // case CppTypeFunction:
        // case CppTypeClass:
        return String("pointer_v");
    }
}


// ---------------------------------------------------------------------------
// Generator::Field::Field(Generator &generator, 
//                         Generator::Class *pContainingClass,
//                         const String &id)
// ---------------------------------------------------------------------------
Generator::Field::Field(Generator &generator, const String &id,
                        Generator::Class *pContainingClass,
                        bool isStatic)
    : Base(generator, id), numberM(fieldNumberG++),
      pContainingClassM(pContainingClass), isStaticM(isStatic)
{
    String bits = pElementM->GetAttributeValue("bits");
    String type = pElementM->GetAttributeValue("type");

    pTypeM = generatorM.GetType(type, bits);
    if (pTypeM == NULL) {
        generatorM.Error
            ("Unable to find field %S type %S\n",
             (const unichar *) pElementM->GetAttributeValue("id"),
             (const unichar *) type);
    }
}


// ---------------------------------------------------------------------------
// Generator::Field::Emit()
// ---------------------------------------------------------------------------
void Generator::Field::Emit()
{
    OutputStream &out = generatorM.GetOutputStream();
    
    // Enusre that the type has been emitted
    pTypeM->Emit();

    // Emit the Get method
    this->EmitGet();

    // Emit the Set method
    this->EmitSet();

    // Emit the Field itself
    out.WriteFA("static XP_Field XrttiPrivate_Field_%lu =\n{\n", 
                  this->GetNumber());

    // access_type
    out.WriteFA("    /* access_type */ %s,\n", this->GetAccessType());

    // name
    out.WriteFA("    /* name */ \"%S\",\n",
                  (const unichar *) this->GetName());

    // type_index
    out.WriteFA("    /* type_index */ %lu,\n", pTypeM->GetNumber());


    // isStatic
    out.WriteFA("    /* is_static */ %s,\n", isStaticM ? "true" : "false");

    // offset
    // statics have no offsetof
    // fields of anonymous classes have no offsetof
    // Bitfields cannot have offsetof
    if (isStaticM || pContainingClassM->IsAnonymous() ||
        (pTypeM->GetCppType() == Type::CppTypeBitfield) ||
        (this->GetName() == "")) {
        out.WriteA("    /* offsetof */ 0,\n");
    }
    else {
        out.WriteFA("    /* offsetof */ XrttiPrivate_offsetof(%S, %S),\n",
                      (const unichar *) pContainingClassM->GetTypedefName(),
                      (const unichar *) this->GetName());
    }

    // get
    out.WriteFA("    /* get */ %S,\n",
                  (const unichar *) this->GetGetMethodName());
    
    // set
    out.WriteFA("    /* set */ %S\n",
                  (const unichar *) this->GetSetMethodName());

    out.WriteA("};\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Field::EmitGet()
// ---------------------------------------------------------------------------
void Generator::Field::EmitGet()
{
    OutputStream &out = generatorM.GetOutputStream();

    out.WriteFA("static void %S(void *pThis, X_Value &returnValue)\n{\n",
                (const unichar *) this->GetGetMethodName());

    // Only if the containing class is not anonymous, and the field itself
    // is not an anonyous struct, do we actually return a value
    if (!pContainingClassM->IsAnonymous() &&
        !((pTypeM->GetCppType() == Type::CppTypeClass) &&
          pTypeM->GetCppTypeClass()->IsAnonymous())) {
        
        const char *deref =  
            ((pTypeM->GetCppType() == Type::CppTypeClass) &&
             !pTypeM->GetPointerDepth()) ? "* " : "";

        const char *amp = 
            ((isStaticM || (pTypeM->GetCppType() == Type::CppTypeBitfield)) ?
             "" : " &");
        
        out.WriteFA
            ("    %s(%S %s&) returnValue.%S = (%S%s) ", deref,
             (const unichar *) pTypeM->GetCppFullType(false),
             deref, (const unichar *) pTypeM->GetValueField(),
             (const unichar *) pTypeM->GetCppFullType(false), amp);
    
        const char *format = 
            (isStaticM ? "%S::%S;\n\n" : "(((%S *) pThis)->%S);\n\n");
        
        out.WriteFA(format, 
                      (const unichar *) pContainingClassM->GetFullName(),
                      (const unichar *) this->GetName());
    }

    out.WriteA("}\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Field::EmitSet()
// ---------------------------------------------------------------------------
void Generator::Field::EmitSet()
{
    OutputStream &out = generatorM.GetOutputStream();
    
    out.WriteFA("static void %S(void *pThis, const X_Value &value)\n{\n    ",
                  (const unichar *) this->GetSetMethodName());

    // A field cannot be set if:
    // The containing class is anonymous, or
    if (pContainingClassM->IsAnonymous() || 
        // The field is const, or
        pTypeM->GetAttributes().isConst ||
        // The field is an anonymous class, or
        ((pTypeM->GetCppType() == Type::CppTypeClass) &&
         pTypeM->GetCppTypeClass()->IsAnonymous()) ||
        // The field is an array
        (pTypeM->GetArrayDimensionsCount() && !pTypeM->GetPointerDepth())) {
        out.WriteA("(void) pThis;\n    (void) value;\n}\n\n");
        return;
    }

    String lhsCast;

    if (pTypeM->GetCppType() != Type::CppTypeBitfield) {
        lhsCast = String::Format
            ("(%S &) ", (const unichar *) pTypeM->GetCppFullType(false));
    }

    const char *lhsFormat = 
        (isStaticM ? "%S%S::%S = " : "%S((%S *) pThis)->%S = ");

    const char *rhsFormat =
        ((pTypeM->GetCppType() == Type::CppTypeClass) &&
         !pTypeM->GetPointerDepth()) ?
        "* (%S *) (value.%S);\n" : "(%S) (value.%S);\n";

    out.WriteFA(lhsFormat, (const unichar *) lhsCast,
                  (const unichar *) pContainingClassM->GetFullName(),
                  (const unichar *) this->GetName());

    out.WriteFA(rhsFormat, (const unichar *) pTypeM->GetCppFullType(false),
                  (const unichar *) pTypeM->GetValueField());

    out.WriteA("}\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Method::Method(Generator &generator, const String &id,
//                           Method::MethodType type)
// ---------------------------------------------------------------------------
Generator::Method::Method(Generator &generator, const String &id,
                          Method::MethodType type)
    : Base(generator, id), methodTypeM(type), hasEllipsisM(false),
      numberM(methodNumberG++)
{
    U32 count = pElementM->GetSubElementCount();

    for (U32 i = 0; i < count; i++) {
        ref<Parser::Element> pSubElement = pElementM->GetSubElement(i);
        
        if (pSubElement->GetElementName() == "Argument") {
            ref<Argument> pArgument = new Argument
                (generatorM, pSubElement->GetAttributeValue("name"),
                 pSubElement->GetAttributeValue("type"),
                 pSubElement->HasAttribute("default"),
                 pSubElement->GetAttributeValue("default"));
            vArgumentsM.Append(pArgument);
        }
        else if (pSubElement->GetElementName() == "Ellipsis") {
            hasEllipsisM = true;
        }
    }

    if ((methodTypeM != MethodTypeConstructor) &&
        (methodTypeM != MethodTypeDestructor)) {
        pReturnTypeM = generator.GetType
            (pElementM->GetAttributeValue("returns"), "");
    }

    String throws = pElementM->GetAttributeValue("throw");

    // Parse the throws
    if (throws != "") {
        RefVector<RefString> vThrows;
        
        split((const unichar *) throws, vThrows);

        RefVector__Enumeration<RefString> enumeration(vThrows);
        
        while (enumeration.HasElements()) {
            String _throw = (RefString *) enumeration.GetNextElement();
            
            // If it's a context, then a type has to be used instead.
            // The thing is, gccxml may not generate a Type for this and
            // instead just point to a context.  In this case, we have to
            // try to find a type which 

            // It's either a Type or a Context.  Try to get it as a type.
            vThrowsM.Append(generator.GetType(_throw, ""));
        }
    }
}


// ---------------------------------------------------------------------------
// Generator::Method::Emit(Class *pContainingClass)
// ---------------------------------------------------------------------------
void Generator::Method::Emit(Class *pContainingClass)
{
    OutputStream &out = generatorM.GetOutputStream();

    // Ensure that the argument types are emitted
    U32 count = this->GetArgumentCount();
    for (U32 i = 0; i < count; i++) {
        this->GetArgument(i)->GetType()->Emit();
    }

    // Ensure that the return type is emitted
    if (pReturnTypeM != NULL) {
        pReturnTypeM->Emit();
    }

    // Now emit the argument names, types, and defaults
    if (count > 0) {
        // names
        
        if (methodTypeM != MethodTypeFunctionSignature) {
            out.WriteFA("static const char * const XrttiPrivate_Method_"
                          "%lu_argument_names[%lu] =\n{\n",
                          this->GetNumber(), count);
            
            for (U32 i = 0; i < count; i++) {
                out.WriteFA("    \"%S\"", (const unichar *) 
                              this->GetArgument(i)->GetName());
                if (i < (count - 1)) {
                    out.WriteA(",\n");
                }
            }
            
            out.WriteA("\n};\n\n");
        }
        
        // types
        out.WriteFA("static const X_U32 XrttiPrivate_Method_%lu_"
                      "argument_types[%lu] =\n{\n", this->GetNumber(), count);
        
        for (U32 i = 0; i < count; i++) {
            out.WriteFA("    %lu", (const unichar *)
                          this->GetArgument(i)->GetType()->GetNumber());
            if (i < (count - 1)) {
                out.WriteA(",\n");
            }
        }
        
        out.WriteA("\n};\n\n");
        
        // defaults
        for (U32 i = 0; i < count; i++) {
            ref<Argument> pArgument = this->GetArgument(i);
            
            if (pArgument->HasDefault()) {
                out.WriteFA("static %S XrttiPrivate_Method_%lu_argument_"
                              "%lu_default = (%S) %S;\n\n",
                              (const unichar *) pArgument->GetType()->
                              GetCppFullType(true),
                              this->GetNumber(), i, 
                              (const unichar *) pArgument->GetType()->
                              GetCppFullType(true),
                              (const unichar *) pArgument->GetDefault());
            }
        }
        
        out.WriteFA("static void *XrttiPrivate_Method_%lu_argument_"
                      "defaults[%lu] =\n{\n", this->GetNumber(), count);

        for (U32 i = 0; i < count; i++) {
            ref<Argument> pArgument = this->GetArgument(i);
            
            if (pArgument->HasDefault()) {
                out.WriteFA("    &XrttiPrivate_Method_%lu_argument_%lu_"
                              "default", this->GetNumber(), i);
            }
            else {
                out.WriteA("    0");
            }
            
            if (i < (count - 1)) {
                out.WriteA(",\n");
            }
        }
        
        out.WriteA("\n};\n\n");
    }

    // Now emit the throws
    count = vThrowsM.Count();
    if (count > 0) {
        out.WriteFA("static X_U32 XrttiPrivate_Method_%lu_throws[%lu] =\n{\n",
                    this->GetNumber(), count);
        for (U32 i = 0; i < count; i++) {
            out.WriteFA("    %lu", vThrowsM[i]->GetNumber());
            if (i < (count - 1)) {
                out.WriteA(",");
            }
            out.WriteA("\n");
        }
        out.WriteA("};\n\n");
    }

    // And emit the invoke method
    if (pContainingClass && (methodTypeM != MethodTypeDestructor)) {
        this->EmitInvoke(pContainingClass);
    }

    // Now the actual method
    out.WriteFA("static XP_Method XrttiPrivate_Method_%lu =\n{\n",
                  this->GetNumber());

    // access_type
    out.WriteFA("    /* access_type */ %s,\n", this->GetAccessType());

    // name
    out.WriteFA("    /* name */ \"%S\",\n",
                  (const unichar *) this->GetName());

    // is_static
    // is_const
    // is_virtual
    // is_pure_virtual
    out.WriteFA("    /* is_static */ %s,\n", 
                  this->IsStatic() ? "true" : "false");
    out.WriteFA("    /* is_const */ %s,\n",
                  this->IsConst() ? "true" : "false");
    out.WriteFA("    /* is_virtual */ %s,\n", 
                  this->IsVirtual() ? "true" : "false");
    out.WriteFA("    /* is_pure_virtual */ %s,\n", 
                  this->IsPureVirtual() ? "true" : "false");

    // return_type_index
    if ((methodTypeM == MethodTypeConstructor) ||
        (methodTypeM == MethodTypeDestructor)) {
        out.WriteA("    /* return_type_index */ 0,\n");
    }
    else {
        out.WriteFA("    /* return_type_index */ %lu,\n",
                      pReturnTypeM->GetNumber());
    }

    // arguments_count
    count = this->GetArgumentCount();
    out.WriteFA("    /* arguments_count */ %lu,\n", count);
                     
    // argument_names
    // argument_types
    // argument_defaults
    if (count) {
        if (methodTypeM == MethodTypeFunctionSignature) {
            out.WriteA("    /* argument_names */ 0,\n");
        }
        else {
            out.WriteFA("    /* argument_names */ XrttiPrivate_Method_%lu_"
                          "argument_names,\n", this->GetNumber());
        }
        out.WriteFA("    /* argument_types */ XrttiPrivate_Method_%lu_"
                      "argument_types,\n", this->GetNumber());
        out.WriteFA("    /* argument_defaults */ XrttiPrivate_Method_%lu_"
                      "argument_defaults,\n", this->GetNumber());
    }
    else {
        out.WriteA("    /* argument_names */ 0,\n");
        out.WriteA("    /* argument_types */ 0,\n");
        out.WriteA("    /* argument_defaults */ 0,\n");
    }
    
    // has_ellipsis
    out.WriteFA("    /* has_ellipsis */ %s,\n", 
                  this->HasEllipsis() ? "true" : "false");

    // throws_count
    // throws
    count = vThrowsM.Count();
    if (count) {
        out.WriteFA("    /* throws_count */ %lu,\n", count);
        out.WriteFA("    /* throws */ XrttiPrivate_Method_%lu_throws,\n",
                    this->GetNumber());
    }
    else {
        out.WriteFA("    /* throws_count */ 0,\n");
        out.WriteFA("    /* throws */ 0,\n");
    }

    // invoke
    if ((pContainingClass == NULL) ||
        (methodTypeM == MethodTypeDestructor)) {
        out.WriteA("    /* invoke */ 0");
    }
    else {
        out.WriteFA("    /* invoke */ & XrttiPrivate_Method_%lu_invoke\n",
                      this->GetNumber());
    }

    out.WriteA("};\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Method::EmitSignature()
// ---------------------------------------------------------------------------
void Generator::Method::EmitSignature()
{
    if (pReturnTypeM == NULL) {
        return;
    }

    OutputStream &out = generatorM.GetOutputStream();

    // Now emit the signature
    out.WriteFA("typedef %S (Method_%lu_signature) (",
                  (const unichar *) pReturnTypeM->GetCppFullType(true),
                  this->GetNumber());

    U32 count = this->GetArgumentCount();

    for (U32 i = 0; i < count; i++) {
        ref<Argument> pArgument = this->GetArgument(i);

        out.WriteA((const unichar *) pArgument->GetType()->
                     GetCppFullType(true));
        
        if (i < (count - 1)) {
            out.WriteA(", ");
        }
    }

    if (this->HasEllipsis()) {
        if (count > 0) {
            out.WriteA(", ");
        }

        out.WriteA("...");
    }
    
    out.WriteA(");\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Method::GetSignature()
// ---------------------------------------------------------------------------
String Generator::Method::GetSignature()
{
    return String::Format("Method_%lu_signature", this->GetNumber());
}


Generator::Type::Type(class Generator &generator, const String &id,
                      const String &bits)
    : Registered(generator, id), bitsM(bits)
{
}

// ---------------------------------------------------------------------------
// Generator::Method::EmitInvoke(Class *pContainingClass)
// ---------------------------------------------------------------------------
void Generator::Method::EmitInvoke(Class *pContainingClass)
{
    OutputStream &out = generatorM.GetOutputStream();

    out.WriteFA("static void XrttiPrivate_Method_%lu_invoke"
                  "(void *pThis, X_Value &returnValue, "
                  "X_Value *pArgumentValues)\n{\n    ", this->GetNumber());

    if ((pContainingClass != NULL) &&
        (pContainingClass->IsAnonymous() || pContainingClass->IsAbstract())) {
        out.WriteA("(void) pThis;\n    (void) returnValue;\n    "
                     "(void) pArgumentValues;");
    }
    else {
        if (methodTypeM == MethodTypeConstructor) {
            out.WriteFA("returnValue.pointer_v = new %S(", (const unichar *)
                          pContainingClass->GetFullName());
        }
        else {
            if (pReturnTypeM->GetCppType() != Type::CppTypeVoid) {
                // If the return value is a pointer, then it can simply
                // be returned, regardless of if it's a reference or not
                const char *pFormat;
                bool useConst = true;
                if (pReturnTypeM->GetPointerDepth()) {
                    pFormat = "(%S &) returnValue.%S = ";
                }
                // Else, the return value is not a pointer; if it's a 
                // class then it has to be handled specially
                else if (pReturnTypeM->GetCppType() == Type::CppTypeClass) {
                    // If it's a reference, then it must be un-ref'd and
                    // returned as a pointer
                    if (pReturnTypeM->IsReference()) {
                        pFormat = "(%S * &) returnValue.%S = & ";
                    }
                    // Otherwise, it's copied
                    else {
                        pFormat = "* (%S *) returnValue.%S = ";
                        useConst = false;
                    }
                }
                // Otherwise, it's a simple type that can be assigned
                // easily
                else {
                    pFormat = "(%S &) returnValue.%S = ";
                }

                out.WriteFA(pFormat, (const unichar *) 
                            pReturnTypeM->GetCppFullType(useConst),
                            (const unichar *) pReturnTypeM->GetValueField());
            }
            
            if (methodTypeM != MethodTypeFunctionSignature) {
                out.WriteFA("((%S *) pThis)->", (const unichar *) 
                              pContainingClass->GetFullName());
            }
            
            out.WriteFA("%s%S(", (methodTypeM == MethodTypeOperatorMethod) ?
                          "operator " : "", (const unichar *) this->GetName());
        }
        
        U32 count = this->GetArgumentCount();
        for (U32 i = 0; i < count; i++) {
            ref<Type> pArgumentType = this->GetArgument(i)->GetType();

            out.WriteFA("(* (%S *) %s(pArgumentValues[%lu].%S))", 
                        (const unichar *) pArgumentType->GetCppFullType(false),
                        ((pArgumentType->GetCppType() != Type::CppTypeClass) ||
                         pArgumentType->GetPointerDepth()) ? "& ": "",
                        i, (const unichar *) pArgumentType->GetValueField());
            
            if (i < (count - 1)) {
                out.WriteA(", ");
            }
        }
        
        out.WriteA(");");
    }

    out.WriteA("\n}\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Class::Initialize()
// ---------------------------------------------------------------------------
void Generator::Class::Initialize()
{
    Context::Initialize();

    String bases = pElementM->GetAttributeValue("bases");

    if (bases != "") {
        RefVector<RefString> vBases;
        
        split((const unichar *) bases, vBases);
        
        RefVector__Enumeration<RefString> enumeration(vBases);
        
        while (enumeration.HasElements()) {
            String base = (RefString *) enumeration.GetNextElement();

            ref<Context> pSuper = generatorM.GetContext(base);

            if (pSuper->GetContextType() != Context::ContextTypeClass) {
                generatorM.Error("Base class of %S is not a class\n",
                                 (const unichar *) this->GetFullName());
            }
            
            vSupersM.Append((Class *) (Context *) pSuper);
        }
    }

    String members = pElementM->GetAttributeValue("members");
    
    RefVector<RefString> vMembers;
    
    split((const unichar *) members, vMembers);
    
    RefVector__Enumeration<RefString> enumeration(vMembers);
    
    while (enumeration.HasElements()) {
        String member = (RefString *) enumeration.GetNextElement();
        
        ref<Parser::Element> pMemberElement =
            generatorM.LookupParserElement(member);
        
        if (pMemberElement == NULL) {
            generatorM.Error("No such member: %S\n",
                             (const unichar *) member);
        }
        
        String memberElementName = pMemberElement->GetElementName();
        
        // If this member is a Field, then add it
        if (memberElementName == "Field") {
            this->AddField(new Field(generatorM, member, this, false));
        }
        // If this member is a Variable, then add it as a static Field
        else if (memberElementName == "Variable") {
            // But skip it if it's artificial
            // And skip it if it's got no name, which gccxml does for no
            // obvious reason sometimes
            if ((pMemberElement->GetAttributeValue("artificial") != "1") &&
                (pMemberElement->GetAttributeValue("name") != "")) {
                this->AddField(new Field(generatorM, member, this, true));
            }
        }
        // Else if it's a Constructor, then add it
        else if (memberElementName == "Constructor") {
            vConstructorsM.Append(new Method(generatorM, member, 
                                             Method::MethodTypeConstructor));
        }
        // Else if it's a Destructor, then add it
        else if (memberElementName == "Destructor") {
            if (pDestructorM) {
                generatorM.Error("Duplicate destructor");
            }
            pDestructorM = new Method(generatorM, member, 
                                      Method::MethodTypeDestructor);
        }
        // Else if it's a Method (or OperatorMethod), then add it
        else if (memberElementName == "Method") {
            vMethodsM.Append(new Method(generatorM, member,
                                        Method::MethodTypeMethod));
        }
        else if (memberElementName == "OperatorMethod") {
            vMethodsM.Append(new Method(generatorM, member,
                                        Method::MethodTypeOperatorMethod));
        }
        // Otherwise, ignore it, we don't care about it
    }
}


// ---------------------------------------------------------------------------
// Generator::Class::EmitTypedef()
// ---------------------------------------------------------------------------
void Generator::Class::EmitTypedef()
{
    generatorM.GetOutputStream().WriteFA
        ("typedef %S %S;\n", (const unichar *) this->GetFullName(),
         (const unichar *) this->GetTypedefName());
}


// ---------------------------------------------------------------------------
// Generator::Class::Emit()
// ---------------------------------------------------------------------------
void Generator::Class::Emit()
{
    if (!this->ShouldEmit()) {
        return;
    }

    OutputStream &out = generatorM.GetOutputStream();
    
    if (!this->IsIncomplete()) {
        // Emit supers array and supers offsets array
        U32 count = vSupersM.Count();
        if (count) {
            out.WriteFA("static const X_U32 XrttiPrivate_Context_%lu_"
                          "supers[%lu] =\n{\n", this->GetNumber(), count);
            
            for (U32 i = 0; i < count; i++) {
                // Add the super to the array
                out.WriteFA("    %lu", vSupersM[i]->GetNumber());
                
                if (i < (count - 1)) {
                    out.WriteA(",\n");
                }
            }
            
            out.WriteA("\n};\n\n");
            
            out.WriteFA("static const X_U32 XrttiPrivate_Context_%lu_"
                          "supers_offsets[%lu] =\n{\n", this->GetNumber(),
                          count);
            
            for (U32 i = 0; i < count; i++) {
                ref<Class> pSuper = vSupersM[i];
                if (this->IsAnonymous()) {
                    out.WriteA("    0");
                }
                else {
                    out.WriteFA("    XrttiPrivate_classoffsetof(%S, %S)",
                                  (const unichar *) pSuper->GetTypedefName(), 
                                  (const unichar *) this->GetTypedefName());
                }
                
                if (i < (count - 1)) {
                    out.WriteA(",\n");
                }
            }
            
            out.WriteA("\n};\n\n");
        }
        
        // Emit the fields for this class
        this->EmitFields();

        // Emit the constructors for this class
        this->EmitConstructors();

        // Emit the destructor for this class (if there is one)
        this->EmitDestructor();
        
        // Emit the methods for this class
        this->EmitMethods();
        
        if (this->CanBeCreatedByDefaultConstructor()) {
            // Emit the create_one function
            out.WriteFA("static void *XrttiPrivate_Context_%lu_create_one()"
                          "\n{\n    return (void *) new %S;\n}\n\n",
                          this->GetNumber(),
                          (const unichar *) this->GetFullName());
            
            // Emit the create_array function
            out.WriteFA("static void *XrttiPrivate_Context_%lu_create_array"
                          "(X_U32 count)\n{\n    return (void *) new %S[count];"
                          "\n}\n\n", this->GetNumber(), 
                          (const unichar *) this->GetFullName());
        }

        if (this->CanBeDestroyedByDefaultDestructor()) {
            // Emit the delete_one function
            out.WriteFA("static void XrttiPrivate_Context_%lu_delete_one"
                          "(void *pInstance)\n{\n    delete (%S *) pInstance;"
                          "\n}\n\n", this->GetNumber(),
                          (const unichar *) this->GetFullName());
            
            // Emit the delete_array function
            out.WriteFA("static void XrttiPrivate_Context_%lu_delete_array"
                          "(void *pInstance)\n{\n    delete [] "
                          "(%S *) pInstance;\n}\n\n", this->GetNumber(),
                          (const unichar *) this->GetFullName());
        }
    }

    // Now emit the context for this class
    out.WriteFA("static const XP_Context XrttiPrivate_Context_%lu =\n{\n",
                  this->GetNumber());

    // type
    out.WriteFA("    /* type */ %s,\n", this->GetContextTypeName());

    // name
    out.WriteFA("    /* name */ \"%S\",\n", 
                  (const unichar *) this->GetName());

    // full
    out.WriteFA("    /* full_name */ \"%S\",\n", 
                  (const unichar *) this->GetFullName());

    // parent
    ref<Context> pParent = this->GetContext();
    if (pParent) {
        out.WriteFA("    /* parent */ %lu", pParent->GetNumber());
    }
    else {
        out.WriteA("    /* parent */ 0xFFFFFFFF");
    }

    // Anonymous and incomplete are truncated here
    if (this->IsIncomplete()) {
        out.WriteA("\n};\n\n");
        return;
    }
    else {
        out.WriteA(",\n");
    }

    // access_type
    out.WriteFA("    /* access_type */ %s,\n", this->GetAccessType());

    // typeinfo
    String fullName;
    if (generatorM.GetRtti() && !this->IsAnonymous() &&
        ((fullName = this->GetFullName()) != "")) {
        out.WriteFA("    /* typeinfo */ & typeid(%S),\n",
                      (const unichar *) fullName);
    }
    else {
        out.WriteA("    /* typeinfo */ 0,\n");
    }

    // supers_count
    U32 count = vSupersM.Count();
    out.WriteFA("    /* supers_count */ %lu,\n", count);
    
    // supers
    if (count) {
        out.WriteFA("    /* supers */ XrttiPrivate_Context_%lu_supers,\n",
                      this->GetNumber());
    }
    else {
        out.WriteA("    /* supers */ 0,\n");
    }

    // supers_offsets
    if (count) {
        out.WriteFA("    /* supers_offsets */ XrttiPrivate_Context_%lu_"
                      "supers_offsets,\n", this->GetNumber());
    }
    else {
        out.WriteA("    /* supers_offsets */ 0,\n");
    }

    // fields_count
    // fields
    this->EmitVector(this->GetFieldCount(), "fields", true, true);

    // is_anonymous
    out.WriteFA("    /* is_anonymous */ %s,\n", 
                this->GetName().StartsWith(".") ? "true" : "false");
        
    // constructors_count
    // constructors
    this->EmitVector(this->GetConstructorCount(), "constructors", true, true);

    if (this->HasDestructor()) {
        out.WriteFA("    /* destructor */ & XrttiPrivate_Method_%lu,\n",
                    this->GetDestructor()->GetNumber());
    }
    else {
        out.WriteFA("    /* destructor */ 0,\n");
    }

    // methods_count
    // methods
    this->EmitVector(this->GetMethodCount(), "methods", true, true);
    
    // create_one
    // create_array
    if (this->CanBeCreatedByDefaultConstructor()) {
        out.WriteFA("    /* create_one */ XrttiPrivate_Context_%lu_"
                      "create_one,\n", this->GetNumber());

        out.WriteFA("    /* create_array */ XrttiPrivate_Context_%lu_"
                      "create_array,\n", this->GetNumber());
    }
    else {
        out.WriteA("    /* create_one */ 0,\n");

        out.WriteA("    /* create_array */ 0,\n");
    }

    // delete_one
    if (this->CanBeDestroyedByDefaultDestructor()) {
        out.WriteFA("    /* delete_one */ XrttiPrivate_Context_%lu_"
                      "delete_one,\n",this->GetNumber());

        // delete_array
        out.WriteFA("    /* delete_array */ XrttiPrivate_Context_%lu_"
                      "delete_array\n", this->GetNumber());
    }
    else {
        out.WriteA("    /* delete_one */ 0,\n");
        out.WriteA("    /* delete_array */ 0\n");
    }

    out.WriteA("};\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Class::IsAbstract()
// ---------------------------------------------------------------------------
bool Generator::Class::IsAbstract()
{
    U32 count = vMethodsM.Count();

    for (U32 i = 0; i < count; i++) {
        if (vMethodsM[i]->IsPureVirtual()) {
            return true;
        }
    }

    return false;
}


// ---------------------------------------------------------------------------
// Generator::Class::CanBeCreatedByDefaultConstructor()
// ---------------------------------------------------------------------------
bool Generator::Class::CanBeCreatedByDefaultConstructor()
{
    if (this->IsAnonymous() || this->IsAbstract() || this->IsIncomplete()) {
        return false;
    }
    
    // Only if the class has a default constructor does this return true
    U32 count = vConstructorsM.Count();
    
    for (U32 i = 0; i < count; i++) {
        if (vConstructorsM[i]->GetArgumentCount() == 0) {
            return true;
        }
    }

    return false;
}


// ---------------------------------------------------------------------------
// Generator::Class::CanBeDestroyedByDefaultDestructor()
// ---------------------------------------------------------------------------
bool Generator::Class::CanBeDestroyedByDefaultDestructor()
{
    return !(this->IsAnonymous() || this->IsIncomplete());
}


// ---------------------------------------------------------------------------
// Generator::Class::EmitFields()
// ---------------------------------------------------------------------------
void Generator::Class::EmitFields()
{
    U32 count = this->GetFieldCount();

    if (count == 0) {
        return;
    }

    // Emit the fields
    for (U32 i = 0; i < count; i++) {
        this->GetField(i)->Emit();
    }

    // Emit the fields array
    OutputStream &out = generatorM.GetOutputStream();

    out.WriteFA("static const XP_Field XrttiPrivate_Context_%lu_"
                  "fields[%lu] =\n{\n", this->GetNumber(), count);

    for (U32 i = 0; i < count; i++) {
        ref<Field> pField = this->GetField(i);
        out.WriteFA("    XrttiPrivate_Field_%lu", pField->GetNumber());
        if (i < (count - 1)) {
            out.WriteA(",\n");
        }
    }

    out.WriteA("\n};\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Class::EmitConstructors()
// ---------------------------------------------------------------------------
void Generator::Class::EmitConstructors()
{
    U32 count = vConstructorsM.Count();

    if (count == 0) {
        return;
    }

    // Emit the constructors
    for (U32 i = 0; i < count; i++) {
        vConstructorsM[i]->Emit(this);
    }

    // Emit the constructors array
    OutputStream &out = generatorM.GetOutputStream();

    out.WriteFA("static const XP_Method XrttiPrivate_Context_%lu_"
                  "constructors[%lu] =\n{\n", this->GetNumber(), count);

    for (U32 i = 0; i < count; i++) {
        ref<Method> pConstructor = vConstructorsM[i];
        out.WriteFA("    XrttiPrivate_Method_%lu", pConstructor->GetNumber());
        if (i < (count - 1)) {
            out.WriteFA(",\n");
        }
    }

    out.WriteA("\n};\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Class::EmitDestructor()
// ---------------------------------------------------------------------------
void Generator::Class::EmitDestructor()
{
    if (!this->HasDestructor()) {
        return;
    }

    this->GetDestructor()->Emit(this);
}


// ---------------------------------------------------------------------------
// Generator::Class::EmitMethods()
// ---------------------------------------------------------------------------
void Generator::Class::EmitMethods()
{
    U32 count = vMethodsM.Count();

    if (count == 0) {
        return;
    }

    // Emit the methods
    for (U32 i = 0; i < count; i++) {
        vMethodsM[i]->Emit(this);
    }
     
    // Emit the methods array
    OutputStream &out = generatorM.GetOutputStream();

    out.WriteFA("static const XP_Method XrttiPrivate_Context_%lu_"
                  "methods[%lu] =\n{\n", this->GetNumber(), count);

    for (U32 i = 0; i < count; i++) {
        ref<Method> pMethod = vMethodsM[i];
        out.WriteFA("    XrttiPrivate_Method_%lu", pMethod->GetNumber());
        if (i < (count - 1)) {
            out.WriteFA(",\n");
        }
    }

    out.WriteA("\n};\n\n");
}


// ---------------------------------------------------------------------------
// Generator::Class::EmitVector(U32 count, const char *name, bool emitCount)
// ---------------------------------------------------------------------------
void Generator::Class::EmitVector(U32 count, const char *name, bool emitCount,
                                  bool emitComma)
{
    OutputStream &out = generatorM.GetOutputStream();
    
    if (count) {
        if (emitCount) {
            out.WriteFA("    /* %s_count */ %lu,\n", name, count);
        }
        out.WriteFA("    /* %s */ XrttiPrivate_Context_%lu_%s%s\n", name,
                      this->GetNumber(), name, emitComma ? "," : "");
    }
    else {
        if (emitCount) {
            out.WriteFA("    /* %s_count */ 0,\n", name);
        }
        out.WriteFA("    /* %s */ 0%s\n", name, emitComma ? "," : "");
    }
}
